# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest
import uuid

from supysonic.db import Folder, Artist, Album, Track, Playlist, User

from .frontendtestbase import FrontendTestBase


class PlaylistTestCase(FrontendTestBase):
    def setUp(self):
        super().setUp()

        folder = Folder.create(name="Root", path="tests/assets", root=True)
        artist = Artist.create(name="Artist!")
        album = Album.create(name="Album!", artist=artist)

        track = Track.create(
            path="tests/assets/23bytes",
            title="23bytes",
            artist=artist,
            album=album,
            folder=folder,
            root_folder=folder,
            duration=2,
            disc=1,
            number=1,
            bitrate=320,
            last_modification=0,
        )

        playlist = Playlist.create(name="Playlist!", user=User.get(name="alice"))
        for _ in range(4):
            playlist.add(track)
        playlist.save()

        self.playlistid = playlist.id

    def test_index(self):
        self._login("alice", "Alic3")
        rv = self.client.get("/playlist")
        self.assertIn("My playlists", rv.data)

    def test_details(self):
        self._login("alice", "Alic3")
        rv = self.client.get("/playlist/string", follow_redirects=True)
        self.assertIn("Invalid", rv.data)
        rv = self.client.get("/playlist/" + str(uuid.uuid4()), follow_redirects=True)
        self.assertIn("Unknown", rv.data)
        rv = self.client.get("/playlist/" + str(self.playlistid))
        self.assertIn("Playlist!", rv.data)
        self.assertIn("23bytes", rv.data)
        self.assertIn("Artist!", rv.data)
        self.assertIn("Album!", rv.data)

    def test_update(self):
        self._login("bob", "B0b")
        rv = self.client.post("/playlist/string", follow_redirects=True)
        self.assertIn("Invalid", rv.data)
        rv = self.client.post("/playlist/" + str(uuid.uuid4()), follow_redirects=True)
        self.assertIn("Unknown", rv.data)
        rv = self.client.post(
            "/playlist/" + str(self.playlistid), follow_redirects=True
        )
        self.assertNotIn("updated", rv.data)
        self.assertIn("not allowed", rv.data)
        self._logout()

        self._login("alice", "Alic3")
        rv = self.client.post(
            "/playlist/" + str(self.playlistid), follow_redirects=True
        )
        self.assertNotIn("updated", rv.data)
        self.assertIn("Missing", rv.data)
        self.assertEqual(Playlist[self.playlistid].name, "Playlist!")

        rv = self.client.post(
            "/playlist/" + str(self.playlistid),
            data={"name": "abc", "public": True},
            follow_redirects=True,
        )
        self.assertIn("updated", rv.data)
        self.assertNotIn("not allowed", rv.data)
        playlist = Playlist[self.playlistid]
        self.assertEqual(playlist.name, "abc")
        self.assertTrue(playlist.public)

    def test_delete(self):
        self._login("bob", "B0b")
        rv = self.client.get("/playlist/del/string", follow_redirects=True)
        self.assertIn("Invalid", rv.data)
        rv = self.client.get(
            "/playlist/del/" + str(uuid.uuid4()), follow_redirects=True
        )
        self.assertIn("Unknown", rv.data)
        rv = self.client.get(
            "/playlist/del/" + str(self.playlistid), follow_redirects=True
        )
        self.assertIn("not allowed", rv.data)
        self.assertEqual(Playlist.select().count(), 1)
        self._logout()

        self._login("alice", "Alic3")
        rv = self.client.get(
            "/playlist/del/" + str(self.playlistid), follow_redirects=True
        )
        self.assertIn("deleted", rv.data)
        self.assertEqual(Playlist.select().count(), 0)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import inspect
import os
import os.path
import shutil
import sys
import tempfile
import unittest

from supysonic.db import release_database
from supysonic.config import DefaultConfig
from supysonic.managers.user import UserManager
from supysonic.web import create_application


class TestConfig(DefaultConfig):
    TESTING = True
    LOGGER_HANDLER_POLICY = "never"
    MIMETYPES = {"mp3": "audio/mpeg", "weirdextension": "application/octet-stream"}
    TRANSCODING = {
        "transcoder_mp3_mp3": "echo -n %srcpath %outrate",
        "transcoder_mp3_rnd": "dd if=/dev/urandom bs=1kB count=52 status=none",
        "decoder_mp3": "echo -n Pushing out some mp3 data...",
        "encoder_cat": "cat -",
        "encoder_md5": "md5sum",
    }

    def __init__(self, with_webui, with_api):
        super().__init__()

        for cls in reversed(inspect.getmro(self.__class__)):
            for attr, value in cls.__dict__.items():
                if attr.startswith("_") or attr != attr.upper():
                    continue

                if isinstance(value, dict):
                    setattr(self, attr, value.copy())
                else:
                    setattr(self, attr, value)

        self.WEBAPP.update({"mount_webui": with_webui, "mount_api": with_api})

        with tempfile.NamedTemporaryFile() as tf:
            if sys.platform == "win32":
                self.DAEMON["socket"] = "\\\\.\\pipe\\" + os.path.basename(tf.name)
            else:
                self.DAEMON["socket"] = tf.name


class MockResponse:
    def __init__(self, response):
        self.__status_code = response.status_code
        self.__data = response.get_data(as_text=True)
        self.__mimetype = response.mimetype

    @property
    def status_code(self):
        return self.__status_code

    @property
    def data(self):
        return self.__data

    @property
    def mimetype(self):
        return self.__mimetype


def patch_method(f):
    original = f

    def patched(*args, **kwargs):
        rv = original(*args, **kwargs)
        return MockResponse(rv)

    return patched


class TestBase(unittest.TestCase):
    __with_webui__ = False
    __with_api__ = False

    def setUp(self):
        self.__db = tempfile.mkstemp()
        self.__dir = tempfile.mkdtemp()
        self.config = TestConfig(self.__with_webui__, self.__with_api__)
        self.config.BASE["database_uri"] = "sqlite:///" + self.__db[1]
        self.config.WEBAPP["cache_dir"] = self.__dir

        self.__app = create_application(self.config)
        self.client = self.__app.test_client()

        UserManager.add("alice", "Alic3", admin=True)
        UserManager.add("bob", "B0b")

    def _patch_client(self):
        self.client.get = patch_method(self.client.get)
        self.client.post = patch_method(self.client.post)

    def app_context(self, *args, **kwargs):
        return self.__app.app_context(*args, **kwargs)

    def request_context(self, *args, **kwargs):
        return self.__app.test_request_context(*args, **kwargs)

    def tearDown(self):
        release_database()
        shutil.rmtree(self.__dir)
        os.close(self.__db[0])
        os.remove(self.__db[1])



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2021-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from abc import ABCMeta, abstractmethod

from ..web import create_application


class BaseServer(metaclass=ABCMeta):
    def __init__(
        self, *, host=None, port=None, socket=None, processes=None, threads=None
    ):
        self._host = host
        self._port = port
        self._socket = socket
        self._processes = processes
        self._threads = threads

    @abstractmethod
    def _build_kwargs(self): ...

    @abstractmethod
    def _run(self, **kwargs): ...

    def _load_app(self):
        return create_application()

    def run(self):
        self._run(**self._build_kwargs())



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest

from supysonic.config import IniConfig


class ConfigTestCase(unittest.TestCase):
    def test_sections(self):
        conf = IniConfig("tests/assets/sample.ini")
        for attr in ("TYPES", "BOOLEANS"):
            self.assertTrue(hasattr(conf, attr))
            self.assertIsInstance(getattr(conf, attr), dict)

    def test_types(self):
        conf = IniConfig("tests/assets/sample.ini")

        self.assertIsInstance(conf.TYPES["float"], float)
        self.assertIsInstance(conf.TYPES["int"], int)
        self.assertIsInstance(conf.TYPES["string"], str)

        for t in ("bool", "switch", "yn"):
            self.assertIsInstance(conf.BOOLEANS[t + "_false"], bool)
            self.assertIsInstance(conf.BOOLEANS[t + "_true"], bool)
            self.assertFalse(conf.BOOLEANS[t + "_false"])
            self.assertTrue(conf.BOOLEANS[t + "_true"])

    def test_no_interpolation(self):
        conf = IniConfig("tests/assets/sample.ini")

        self.assertEqual(conf.ISSUE84["variable"], "value")
        self.assertEqual(conf.ISSUE84["key"], "some value with a %variable")


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

import hashlib
import random
import string
import uuid

from ..db import User


class UserManager:
    @staticmethod
    def get(uid):
        if isinstance(uid, uuid.UUID):
            pass
        elif isinstance(uid, str):
            uid = uuid.UUID(uid)
        else:
            raise TypeError("Invalid user id")

        return User[uid]

    @staticmethod
    def add(name, password, **kwargs):
        if User.select().where(User.name == name).exists():
            raise ValueError(f"User '{name}' exists")

        crypt, salt = UserManager.__encrypt_password(password)
        return User.create(name=name, password=crypt, salt=salt, **kwargs)

    @staticmethod
    def delete(uid):
        user = UserManager.get(uid)
        user.delete_instance(recursive=True)

    @staticmethod
    def delete_by_name(name):
        user = User.get(name=name)
        user.delete_instance(recursive=True)

    @staticmethod
    def try_auth(name, password):
        user = User.get_or_none(name=name)
        if user is None:
            return None
        elif UserManager.__encrypt_password(password, user.salt)[0] != user.password:
            return None
        else:
            return user

    @staticmethod
    def change_password(uid, old_pass, new_pass):
        user = UserManager.get(uid)
        if UserManager.__encrypt_password(old_pass, user.salt)[0] != user.password:
            raise ValueError("Wrong password")

        user.password = UserManager.__encrypt_password(new_pass, user.salt)[0]
        user.save()

    @staticmethod
    def change_password2(name_or_user, new_pass):
        if isinstance(name_or_user, User):
            user = name_or_user
        elif isinstance(name_or_user, str):
            user = User.get(name=name_or_user)
        else:
            raise TypeError("Requires a User instance or a user name (string)")

        user.password = UserManager.__encrypt_password(new_pass, user.salt)[0]
        user.save()

    @staticmethod
    def __encrypt_password(password, salt=None):
        if salt is None:
            salt = "".join(random.choice(string.printable.strip()) for _ in range(6))
        return (
            hashlib.sha1(salt.encode("utf-8") + password.encode("utf-8")).hexdigest(),
            salt,
        )



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

# Converts ids from hex-encoded strings to binary data

import argparse

try:
    import MySQLdb as provider
except ImportError:
    import pymysql as provider

from uuid import UUID
from warnings import filterwarnings

parser = argparse.ArgumentParser()
parser.add_argument("username")
parser.add_argument("password")
parser.add_argument("database")
parser.add_argument("-H", "--host", default="localhost", help="default: localhost")
args = parser.parse_args()


def process_table(connection, table, fields, nullable_fields=()):
    to_update = {field: set() for field in fields + nullable_fields}

    c = connection.cursor()
    c.execute("SELECT {1} FROM {0}".format(table, ",".join(fields + nullable_fields)))
    for row in c:
        for field, value in zip(fields + nullable_fields, row):
            if value is None or not isinstance(value, str):
                continue
            to_update[field].add(value)

    for field, values in to_update.iteritems():
        if not values:
            continue
        sql = "UPDATE {0} SET {1}=%s WHERE {1}=%s".format(table, field)
        c.executemany(sql, map(lambda v: (UUID(v).bytes, v), values))
    for field in fields:
        sql = "ALTER TABLE {} MODIFY {} BINARY(16) NOT NULL".format(table, field)
        c.execute(sql)
    for field in nullable_fields:
        sql = "ALTER TABLE {} MODIFY {} BINARY(16)".format(table, field)
        c.execute(sql)

    connection.commit()


filterwarnings("ignore", category=provider.Warning)
conn = provider.connect(
    host=args.host, user=args.username, passwd=args.password, db=args.database
)
conn.cursor().execute("SET FOREIGN_KEY_CHECKS = 0")

process_table(conn, "folder", ("id",), ("parent_id",))
process_table(conn, "artist", ("id",))
process_table(conn, "album", ("id", "artist_id"))
process_table(
    conn, "track", ("id", "album_id", "artist_id", "root_folder_id", "folder_id")
)
process_table(conn, "user", ("id",), ("last_play_id",))
process_table(conn, "client_prefs", ("user_id",))
process_table(conn, "starred_folder", ("user_id", "starred_id"))
process_table(conn, "starred_artist", ("user_id", "starred_id"))
process_table(conn, "starred_album", ("user_id", "starred_id"))
process_table(conn, "starred_track", ("user_id", "starred_id"))
process_table(conn, "rating_folder", ("user_id", "rated_id"))
process_table(conn, "rating_track", ("user_id", "rated_id"))
process_table(conn, "chat_message", ("id", "user_id"))
process_table(conn, "playlist", ("id", "user_id"))

conn.cursor().execute("SET FOREIGN_KEY_CHECKS = 1")
conn.close()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

from supysonic import db
from supysonic.managers.user import UserManager

import unittest
import uuid


class UserManagerTestCase(unittest.TestCase):
    def setUp(self):
        # Create an empty sqlite database in memory
        db.init_database("sqlite:")

    def tearDown(self):
        db.release_database()

    def create_data(self):
        # Create some users
        alice = UserManager.add("alice", "ALICE", admin=True)
        self.assertIsInstance(alice, db.User)
        self.assertTrue(alice.admin)

        bob = UserManager.add("bob", "BOB")
        self.assertIsInstance(bob, db.User)
        self.assertFalse(bob.admin)

        self.assertIsInstance(UserManager.add("charlie", "CHARLIE"), db.User)

        folder = db.Folder.create(name="Root", path="tests/assets", root=True)
        artist = db.Artist.create(name="Artist")
        album = db.Album.create(name="Album", artist=artist)
        track = db.Track.create(
            title="Track",
            disc=1,
            number=1,
            duration=1,
            artist=artist,
            album=album,
            path="tests/assets/empty",
            folder=folder,
            root_folder=folder,
            bitrate=320,
            last_modification=0,
        )

        playlist = db.Playlist(name="Playlist", user=db.User.get(name="alice"))
        playlist.add(track)

    def test_encrypt_password(self):
        func = UserManager._UserManager__encrypt_password
        self.assertEqual(
            func("password", "salt"),
            ("59b3e8d637cf97edbe2384cf59cb7453dfe30789", "salt"),
        )
        self.assertEqual(
            func("pass-word", "pepper"),
            ("d68c95a91ed7773aa57c7c044d2309a5bf1da2e7", "pepper"),
        )
        self.assertEqual(
            func("éèàïô", "ABC+"), ("b639ba5217b89c906019d89d5816b407d8730898", "ABC+")
        )

    def test_get_user(self):
        self.create_data()

        # Get existing users
        for name in ["alice", "bob", "charlie"]:
            user = db.User.get(name=name)
            self.assertEqual(UserManager.get(user.id), user)

        # Get with invalid UUID
        self.assertRaises(ValueError, UserManager.get, "invalid-uuid")
        self.assertRaises(TypeError, UserManager.get, 0xFEE1BAD)

        # Non-existent user
        self.assertRaises(db.User.DoesNotExist, UserManager.get, uuid.uuid4())

    def test_add_user(self):
        self.create_data()
        self.assertEqual(db.User.select().count(), 3)

        # Create duplicate
        self.assertRaises(ValueError, UserManager.add, "alice", "Alic3", admin=True)

    def test_delete_user(self):
        self.create_data()

        # Delete invalid UUID
        self.assertRaises(ValueError, UserManager.delete, "invalid-uuid")
        self.assertRaises(TypeError, UserManager.delete, 0xFEE1B4D)
        self.assertEqual(db.User.select().count(), 3)

        # Delete non-existent user
        self.assertRaises(db.User.DoesNotExist, UserManager.delete, uuid.uuid4())
        self.assertEqual(db.User.select().count(), 3)

        # Delete existing users
        for name in ["alice", "bob", "charlie"]:
            user = db.User.get(name=name)
            db.ClientPrefs.create(
                user=user, client_name="tests"
            )  # test for FK handling
            UserManager.delete(user.id)
            self.assertRaises(db.User.DoesNotExist, db.User.__getitem__, user.id)
        self.assertEqual(db.User.select().count(), 0)

    def test_delete_by_name(self):
        self.create_data()

        # Delete existing users
        for name in ["alice", "bob", "charlie"]:
            user = db.User.get(name=name)
            db.ClientPrefs.create(
                user=user, client_name="tests"
            )  # test for FK handling
            UserManager.delete_by_name(name)
            self.assertFalse(db.User.select().where(db.User.name == name).exists())

        # Delete non-existent user
        self.assertRaises(db.User.DoesNotExist, UserManager.delete_by_name, "null")

    def test_try_auth(self):
        self.create_data()

        # Test authentication
        for name in ["alice", "bob", "charlie"]:
            user = db.User.get(name=name)
            authed = UserManager.try_auth(name, name.upper())
            self.assertEqual(authed, user)

        # Wrong password
        self.assertIsNone(UserManager.try_auth("alice", "bad"))
        self.assertIsNone(UserManager.try_auth("alice", "alice"))

        # Non-existent user
        self.assertIsNone(UserManager.try_auth("null", "null"))

    def test_change_password(self):
        self.create_data()

        # With existing users
        for name in ["alice", "bob", "charlie"]:
            user = db.User.get(name=name)
            # Good password
            UserManager.change_password(user.id, name.upper(), "newpass")
            self.assertEqual(UserManager.try_auth(name, "newpass"), user)
            # Old password
            self.assertEqual(UserManager.try_auth(name, name.upper()), None)
            # Wrong password
            self.assertRaises(
                ValueError, UserManager.change_password, user.id, "badpass", "newpass"
            )

        # Ensure we still got the same number of users
        self.assertEqual(db.User.select().count(), 3)

        # With invalid UUID
        self.assertRaises(
            ValueError,
            UserManager.change_password,
            "invalid-uuid",
            "oldpass",
            "newpass",
        )

        # Non-existent user
        self.assertRaises(
            db.User.DoesNotExist,
            UserManager.change_password,
            uuid.uuid4(),
            "oldpass",
            "newpass",
        )

    def test_change_password2(self):
        self.create_data()

        self.assertRaises(TypeError, UserManager.change_password2, uuid.uuid4(), "pass")

        # With existing users
        for name in ["alice", "bob", "charlie"]:
            UserManager.change_password2(name, "newpass")
            user = db.User.get(name=name)
            self.assertEqual(UserManager.try_auth(name, "newpass"), user)
            self.assertEqual(UserManager.try_auth(name, name.upper()), None)

            # test passing the user directly
            UserManager.change_password2(user, "NEWPASS")
            self.assertEqual(UserManager.try_auth(name, "NEWPASS"), user)

        # Non-existent user
        self.assertRaises(
            db.User.DoesNotExist, UserManager.change_password2, "null", "newpass"
        )


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2019-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import os.path
import shutil
import sys
import tempfile
import unittest

from supysonic.db import init_database, release_database
from supysonic.managers.folder import FolderManager
from supysonic.scanner import Scanner


@unittest.skipIf(sys.platform == "win32", "Windows doesn't allow space-only filenames")
class Issue148TestCase(unittest.TestCase):
    def setUp(self):
        self.__dir = tempfile.mkdtemp()
        init_database("sqlite:")
        FolderManager.add("folder", self.__dir)

    def tearDown(self):
        release_database()
        shutil.rmtree(self.__dir)

    def test_issue(self):
        subdir = os.path.join(self.__dir, "  ")
        os.makedirs(subdir)
        shutil.copyfile(
            "tests/assets/folder/silence.mp3", os.path.join(subdir, "silence.mp3")
        )

        scanner = Scanner()
        scanner.queue_folder("folder")
        scanner.run()


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import uuid

from flask import request

from ..db import Playlist, User, Track, db

from . import get_entity, api_routing
from .exceptions import Forbidden, MissingParameter


@api_routing("/getPlaylists")
def list_playlists():
    query = (
        Playlist.select()
        .orwhere(Playlist.user == request.user, Playlist.public)
        .order_by(Playlist.name)
    )

    username = request.values.get("username")
    if username:
        if not request.user.admin:
            raise Forbidden()

        # get rather than join in the following query to raise an exception if the
        # requested user doesn't exist
        user = User.get(name=username)
        query = Playlist.select().where(Playlist.user == user).order_by(Playlist.name)

    return request.formatter(
        "playlists",
        {"playlist": [p.as_subsonic_playlist(request.user) for p in query]},
    )


@api_routing("/getPlaylist")
def show_playlist():
    res = get_entity(Playlist)
    if res.user != request.user and not res.public and not request.user.admin:
        raise Forbidden()

    info = res.as_subsonic_playlist(request.user)
    info["entry"] = [
        t.as_subsonic_child(request.user, request.client) for t in res.get_tracks()
    ]
    return request.formatter("playlist", info)


@api_routing("/createPlaylist")
@db.atomic()
def create_playlist():
    playlist_id, name = map(request.values.get, ("playlistId", "name"))
    # songId actually doesn't seem to be required
    songs = request.values.getlist("songId")
    playlist_id = uuid.UUID(playlist_id) if playlist_id else None

    if playlist_id:
        playlist = Playlist[playlist_id]

        if playlist.user != request.user and not request.user.admin:
            raise Forbidden()

        playlist.clear()
        if name:
            playlist.name = name
    elif name:
        playlist = Playlist.create(user=request.user, name=name)
    else:
        raise MissingParameter("playlistId or name")

    for sid in songs:
        sid = uuid.UUID(sid)
        track = Track[sid]
        playlist.add(track)
    playlist.save()

    return request.formatter.empty


@api_routing("/deletePlaylist")
def delete_playlist():
    res = get_entity(Playlist)
    if res.user != request.user and not request.user.admin:
        raise Forbidden()

    res.delete_instance()
    return request.formatter.empty


@api_routing("/updatePlaylist")
def update_playlist():
    res = get_entity(Playlist, "playlistId")
    if res.user != request.user and not request.user.admin:
        raise Forbidden()

    playlist = res
    name, comment, public = map(request.values.get, ("name", "comment", "public"))
    to_add, to_remove = map(
        request.values.getlist, ("songIdToAdd", "songIndexToRemove")
    )

    if name:
        playlist.name = name
    if comment:
        playlist.comment = comment
    if public:
        playlist.public = public in (True, "True", "true", 1, "1")

    to_add = map(uuid.UUID, to_add)
    to_remove = map(int, to_remove)

    for sid in to_add:
        track = Track[sid]
        playlist.add(track)

    playlist.remove_at_indexes(to_remove)
    playlist.save()

    return request.formatter.empty



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

# Converts ids from binary data to hex-encoded strings

import sqlite3

from uuid import UUID


def process_table(connection, table, fields):
    to_update = {field: set() for field in fields}

    c = connection.cursor()
    for row in c.execute("SELECT {1} FROM {0}".format(table, ",".join(fields))):
        for field, value in zip(fields, row):
            if value is None or not isinstance(value, bytes):
                continue
            to_update[field].add(value)

    for field, values in to_update.items():
        sql = "UPDATE {0} SET {1}=? WHERE {1}=?".format(table, field)
        c.executemany(sql, map(lambda v: (UUID(bytes=v).hex, v), values))

    connection.commit()


def apply(args):
    file = args.pop("database")
    with sqlite3.connect(file, **args) as conn:
        conn.cursor().execute("PRAGMA foreign_keys = OFF")

        process_table(conn, "artist", ("id",))
        process_table(conn, "album", ("id", "artist_id"))
        process_table(conn, "track", ("id", "album_id", "artist_id"))
        process_table(conn, "user", ("id", "last_play_id"))
        process_table(conn, "client_prefs", ("user_id",))
        process_table(conn, "starred_folder", ("user_id",))
        process_table(conn, "starred_artist", ("user_id", "starred_id"))
        process_table(conn, "starred_album", ("user_id", "starred_id"))
        process_table(conn, "starred_track", ("user_id", "starred_id"))
        process_table(conn, "rating_folder", ("user_id",))
        process_table(conn, "rating_track", ("user_id", "rated_id"))
        process_table(conn, "chat_message", ("id", "user_id"))
        process_table(conn, "playlist", ("id", "user_id"))
        process_table(conn, "radio_station", ("id",))



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import os.path

from ..daemon.client import DaemonClient
from ..daemon.exceptions import DaemonUnavailableError
from ..db import Folder, Artist, Album


class FolderManager:
    @staticmethod
    def get(id):
        try:
            id = int(id)
        except ValueError:
            raise ValueError("Invalid folder id")

        return Folder[id]

    @staticmethod
    def add(name, path):
        try:
            Folder.get(name=name, root=True)
            raise ValueError(f"Folder '{name}' exists")
        except Folder.DoesNotExist:
            pass

        path = os.path.abspath(os.path.expanduser(path))
        if not os.path.isdir(path):
            raise ValueError("The path doesn't exits or isn't a directory")

        try:
            Folder.get(path=path)
            raise ValueError("This path is already registered")
        except Folder.DoesNotExist:
            pass

        if any(
            path.startswith(p)
            for (p,) in Folder.select(Folder.path).where(Folder.root).tuples()
        ):
            raise ValueError("This path is already registered")
        if Folder.select().where(Folder.path.startswith(path)).exists():
            raise ValueError("This path contains a folder that is already registered")

        folder = Folder.create(root=True, name=name, path=path)
        try:
            DaemonClient().add_watched_folder(path)
        except DaemonUnavailableError:
            pass

        return folder

    @staticmethod
    def delete(id):
        folder = FolderManager.get(id)
        if not folder.root:
            raise Folder.DoesNotExist(id)

        try:
            DaemonClient().remove_watched_folder(folder.path)
        except DaemonUnavailableError:
            pass

        folder.delete_hierarchy()
        Album.prune()
        Artist.prune()

    @staticmethod
    def delete_by_name(name):
        folder = Folder.get(name=name, root=True)
        FolderManager.delete(folder.id)



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import mutagen
import os
import os.path
import shutil
import tempfile
import unittest

from contextlib import contextmanager

from supysonic import db
from supysonic.managers.folder import FolderManager
from supysonic.scanner import Scanner


class ScannerTestCase(unittest.TestCase):
    def setUp(self):
        db.init_database("sqlite:")

        folder = FolderManager.add("folder", os.path.abspath("tests/assets/folder"))
        self.assertIsNotNone(folder)

        self.folderid = folder.id
        self.__scan()

    def tearDown(self):
        db.release_database()

    @contextmanager
    def __temporary_track_copy(self):
        track = db.Track.select().first()
        with tempfile.NamedTemporaryFile(
            dir=os.path.dirname(track.path), delete=False
        ) as tf:
            with open(track.path, "rb") as f:
                tf.write(f.read())
        try:
            yield tf.name
        finally:
            os.remove(tf.name)

    def __scan(self, force=False):
        self.scanner = Scanner(force=force)
        self.scanner.queue_folder("folder")
        self.scanner.run()

    def test_scan(self):
        self.assertEqual(db.Track.select().count(), 1)

        self.assertRaises(TypeError, self.scanner.queue_folder, None)
        self.assertRaises(
            TypeError, self.scanner.queue_folder, db.Folder[self.folderid]
        )

    def test_rescan(self):
        self.__scan()
        self.assertEqual(db.Track.select().count(), 1)

    def test_force_rescan(self):
        self.__scan(True)
        self.assertEqual(db.Track.select().count(), 1)

    def test_scan_file(self):
        self.scanner.scan_file("/some/inexistent/path")
        self.assertEqual(db.Track.select().count(), 1)

    def test_scanned_metadata(self):
        self.assertEqual(db.Track.select().count(), 1)

        track = db.Track.select().first()
        artist = db.Artist.select().where(db.Artist.id == track.artist).first()
        album = db.Album.select().where(db.Album.id == track.album).first()

        self.assertEqual(track.bitrate, 128)
        self.assertEqual(track.disc, 1)
        self.assertEqual(track.number, 1)
        self.assertEqual(track.duration, 4)
        self.assertEqual(track.has_art, True)
        self.assertEqual(track.title, "[silence]")
        self.assertEqual(artist.name, "Some artist")
        self.assertEqual(album.name, "Awesome album")

    def test_remove_file(self):
        track = db.Track.select().first()
        self.assertRaises(TypeError, self.scanner.remove_file, None)
        self.assertRaises(TypeError, self.scanner.remove_file, track)

        self.scanner.remove_file("/some/inexistent/path")
        self.assertEqual(db.Track.select().count(), 1)

        self.scanner.remove_file(track.path)
        self.scanner.prune()
        self.assertEqual(db.Track.select().count(), 0)
        self.assertEqual(db.Album.select().count(), 0)
        self.assertEqual(db.Artist.select().count(), 0)

    def test_move_file(self):
        track = db.Track.select().first()
        self.assertRaises(TypeError, self.scanner.move_file, None, "string")
        self.assertRaises(TypeError, self.scanner.move_file, track, "string")
        self.assertRaises(TypeError, self.scanner.move_file, "string", None)
        self.assertRaises(TypeError, self.scanner.move_file, "string", track)

        self.scanner.move_file("/some/inexistent/path", track.path)
        self.assertEqual(db.Track.select().count(), 1)

        self.scanner.move_file(track.path, track.path)
        self.assertEqual(db.Track.select().count(), 1)

        self.assertRaises(
            Exception, self.scanner.move_file, track.path, "/some/inexistent/path"
        )

        with self.__temporary_track_copy() as tf:
            self.__scan()
            self.assertEqual(db.Track.select().count(), 2)
            self.scanner.move_file(tf, track.path)
            self.assertEqual(db.Track.select().count(), 1)

        track = db.Track.select().first()
        new_path = track.path.replace("silence", "silence_moved")
        self.scanner.move_file(track.path, new_path)

        track = db.Track.select().first()
        self.assertEqual(db.Track.select().count(), 1)
        self.assertEqual(track.path, new_path)

    def test_rescan_corrupt_file(self):
        with self.__temporary_track_copy() as tf:
            self.__scan()
            self.assertEqual(db.Track.select().count(), 2)

            with open(tf, "wb") as f:
                f.seek(0, 0)
                f.write(b"\x00" * 4096)
                f.truncate()

            self.__scan(True)
            self.assertEqual(db.Track.select().count(), 1)

    def test_rescan_removed_file(self):
        with self.__temporary_track_copy():
            self.__scan()
            self.assertEqual(db.Track.select().count(), 2)

        self.__scan()
        self.assertEqual(db.Track.select().count(), 1)

    def test_scan_tag_change(self):
        with self.__temporary_track_copy() as tf:
            self.__scan()
            copy = db.Track.get(path=tf)
            self.assertEqual(copy.artist.name, "Some artist")
            self.assertEqual(copy.album.name, "Awesome album")

            tags = mutagen.File(copy.path, easy=True)
            tags["artist"] = "Renamed artist"
            tags["album"] = "Crappy album"
            tags.save()

            self.__scan(True)
            copy = db.Track.get(path=tf)
            self.assertEqual(copy.artist.name, "Renamed artist")
            self.assertEqual(copy.album.name, "Crappy album")
            self.assertIsNotNone(db.Artist.get(name="Some artist"))
            self.assertIsNotNone(db.Album.get(name="Awesome album"))

    def test_stats(self):
        stats = self.scanner.stats()
        self.assertEqual(stats.added.artists, 1)
        self.assertEqual(stats.added.albums, 1)
        self.assertEqual(stats.added.tracks, 1)
        self.assertEqual(stats.deleted.artists, 0)
        self.assertEqual(stats.deleted.albums, 0)
        self.assertEqual(stats.deleted.tracks, 0)


class ScannerDeletionsTestCase(unittest.TestCase):
    def setUp(self):
        self.__dir = tempfile.mkdtemp()
        db.init_database("sqlite:")
        FolderManager.add("folder", self.__dir)

        # Create folder hierarchy
        self._firstsubdir = tempfile.mkdtemp(dir=self.__dir)
        subdir = self._firstsubdir
        for _ in range(4):
            subdir = tempfile.mkdtemp(dir=subdir)

        # Put a file in the deepest folder
        self._trackpath = os.path.join(subdir, "silence.mp3")
        shutil.copyfile("tests/assets/folder/silence.mp3", self._trackpath)

        self._scan()

        # Create annotation data
        track = db.Track.get()
        firstdir = db.Folder.get(path=self._firstsubdir)
        user = db.User.create(
            name="user", password="password", salt="salt", last_play=track
        )
        db.StarredFolder.create(user=user, starred=track.folder_id)
        db.StarredFolder.create(user=user, starred=firstdir)
        db.StarredArtist.create(user=user, starred=track.artist_id)
        db.StarredAlbum.create(user=user, starred=track.album_id)
        db.StarredTrack.create(user=user, starred=track)
        db.RatingFolder.create(user=user, rated=track.folder_id, rating=2)
        db.RatingFolder.create(user=user, rated=firstdir, rating=2)
        db.RatingTrack.create(user=user, rated=track, rating=2)

    def tearDown(self):
        db.release_database()
        shutil.rmtree(self.__dir)

    def _scan(self):
        scanner = Scanner()
        scanner.queue_folder("folder")
        scanner.run()

        return scanner.stats()

    def _check_assertions(self, stats):
        self.assertEqual(stats.deleted.artists, 1)
        self.assertEqual(stats.deleted.albums, 1)
        self.assertEqual(stats.deleted.tracks, 1)
        self.assertEqual(db.Track.select().count(), 0)
        self.assertEqual(db.Album.select().count(), 0)
        self.assertEqual(db.Artist.select().count(), 0)
        self.assertEqual(db.User.select().count(), 1)
        self.assertEqual(db.Folder.select().count(), 1)

    def test_parent_folder(self):
        shutil.rmtree(self._firstsubdir)
        stats = self._scan()
        self._check_assertions(stats)

    def test_track(self):
        os.remove(self._trackpath)
        stats = self._scan()
        self._check_assertions(stats)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

# Converts ids from binary data to hex-encoded strings

try:
    import MySQLdb as provider
except ImportError:
    import pymysql as provider

from uuid import UUID
from warnings import filterwarnings


def process_table(connection, table, fields, nullable_fields=()):
    to_update = {field: set() for field in fields + nullable_fields}

    c = connection.cursor()
    c.execute("SELECT {1} FROM {0}".format(table, ",".join(fields + nullable_fields)))
    for row in c:
        for field, value in zip(fields + nullable_fields, row):
            if value is None or not isinstance(value, bytes):
                continue
            to_update[field].add(value)

    for field in fields:
        sql = "ALTER TABLE {} MODIFY {} BINARY(32) NOT NULL".format(table, field)
        c.execute(sql)
    for field in nullable_fields:
        sql = "ALTER TABLE {} MODIFY {} BINARY(32)".format(table, field)
        c.execute(sql)
    for field, values in to_update.items():
        if not values:
            continue
        sql = "UPDATE {0} SET {1}=%s WHERE {1}=%s".format(table, field)
        c.executemany(
            sql, map(lambda v: (UUID(bytes=v).hex, v + (b"\x00" * 16)), values)
        )
    for field in fields:
        sql = "ALTER TABLE {} MODIFY {} CHAR(32) NOT NULL".format(table, field)
        c.execute(sql)
    for field in nullable_fields:
        sql = "ALTER TABLE {} MODIFY {} CHAR(32)".format(table, field)
        c.execute(sql)

    connection.commit()


def apply(args):
    filterwarnings("ignore", category=provider.Warning)

    conn = provider.connect(**args)
    conn.cursor().execute("SET FOREIGN_KEY_CHECKS = 0")

    process_table(conn, "artist", ("id",))
    process_table(conn, "album", ("id", "artist_id"))
    process_table(conn, "track", ("id", "album_id", "artist_id"))
    process_table(conn, "user", ("id",), ("last_play_id",))
    process_table(conn, "client_prefs", ("user_id",))
    process_table(conn, "starred_folder", ("user_id",))
    process_table(conn, "starred_artist", ("user_id", "starred_id"))
    process_table(conn, "starred_album", ("user_id", "starred_id"))
    process_table(conn, "starred_track", ("user_id", "starred_id"))
    process_table(conn, "rating_folder", ("user_id",))
    process_table(conn, "rating_track", ("user_id", "rated_id"))
    process_table(conn, "chat_message", ("id", "user_id"))
    process_table(conn, "playlist", ("id", "user_id"))
    process_table(conn, "radio_station", ("id",))

    conn.cursor().execute("SET FOREIGN_KEY_CHECKS = 1")
    conn.close()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2024 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import re
import logging

from lxml import etree

from ..testbase import TestBase

path_replace_regexp = re.compile(r"/(\w+)")

NS = "http://subsonic.org/restapi"
NSMAP = {"sub": NS}


class ApiTestBase(TestBase):
    __with_api__ = True

    def setUp(self, apiVersion="1.12.0"):
        super().setUp()
        logging.getLogger("supysonic.api").addHandler(logging.NullHandler())
        self.apiVersion = apiVersion
        xsd = etree.parse(f"tests/assets/subsonic-rest-api-{self.apiVersion}.xsd")
        self.schema = etree.XMLSchema(xsd)

    def _find(self, xml, path):
        """
        Helper method that insert the namespace in ElementPath 'path'
        """

        path = path_replace_regexp.sub(rf"/{{{NS}}}\1", path)
        return xml.find(path)

    def _xpath(self, elem, path):
        """
        Helper method that insert a prefix and map the namespace in XPath 'path'
        """

        path = path_replace_regexp.sub(r"/sub:\1", path)
        return elem.xpath(path, namespaces=NSMAP)

    def _make_request(self, endpoint, args={}, tag=None, error=None, skip_post=False):
        """
        Makes both a GET and POST requests against the API, assert both get the same response.
        If the user isn't provided with the 'u' and 'p' in 'args', the default 'alice' is used.
        Validate the response against the Subsonic API XSD and assert it matches the expected tag or error.

        :param endpoint: request endpoint, with the '/rest/'prefix nor the '.view' suffix
        :param args: dict of parameters. 'u', 'p', 'c' and 'v' are automatically added
        :param tag: topmost expected element, right beneath 'subsonic-response'
        :param error: if the given 'args' should produce an error at 'endpoint', this is the expected error code
        :param skip_post: don't do the POST request

        :return: a 2-tuple (resp, child) if no error, where 'resp' is the full response object, 'child' a
            'lxml.etree.Element' mathching 'tag' (if any). If there's an error (when expected), only returns
            the response object
        """

        if not isinstance(args, dict):
            raise TypeError("'args', expecting a dict, got " + type(args).__name__)
        if tag and not isinstance(tag, str):
            raise TypeError("'tag', expecting a str, got " + type(tag).__name__)

        args.update({"c": "tests", "v": self.apiVersion})
        if "u" not in args:
            args.update({"u": "alice", "p": "Alic3"})

        uri = f"/rest/{endpoint}.view"
        rg = self.client.get(uri, query_string=args)
        if not skip_post:
            rp = self.client.post(uri, data=args)
            self.assertEqual(rg.data, rp.data)

        xml = etree.fromstring(rg.data)
        self.schema.assert_(xml)

        if xml.get("status") == "ok":
            self.assertIsNone(error)
            if tag:
                self.assertEqual(xml[0].tag, f"{{{NS}}}{tag}")
                return rg, xml[0]
            else:
                self.assertEqual(len(xml), 0)
                return rg, None
        else:
            self.assertIsNone(tag)
            self.assertEqual(xml[0].tag, f"{{{NS}}}error")
            self.assertEqual(xml[0].get("code"), str(error))
            return rg



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import time
import unittest
import uuid

from supysonic.db import Folder, Artist, Album, Track

from .apitestbase import ApiTestBase


class BrowseTestCase(ApiTestBase):
    def setUp(self):
        super().setUp()

        self.empty_root = Folder.create(root=True, name="Empty root", path="/tmp")
        self.root = Folder.create(root=True, name="Root folder", path="tests/assets")

        for letter in "ABC":
            folder = Folder.create(
                name=letter + "rtist",
                path=f"tests/assets/{letter}rtist",
                root=False,
                parent=self.root,
            )

            artist = Artist.create(name=letter + "rtist")

            for lether in "AB":
                afolder = Folder.create(
                    name=letter + lether + "lbum",
                    path="tests/assets/{0}rtist/{0}{1}lbum".format(letter, lether),
                    root=False,
                    parent=folder,
                )

                album = Album.create(name=letter + lether + "lbum", artist=artist)

                for num, song in enumerate(["One", "Two", "Three"]):
                    Track.create(
                        disc=1,
                        number=num,
                        title=song,
                        duration=2,
                        album=album,
                        artist=artist,
                        genre="Music!",
                        bitrate=320,
                        path="tests/assets/{0}rtist/{0}{1}lbum/{2}".format(
                            letter, lether, song
                        ),
                        last_modification=0,
                        root_folder=self.root,
                        folder=afolder,
                    )

        self.assertEqual(Folder.select().count(), 11)
        self.assertEqual(Folder.select().where(Folder.root).count(), 2)
        self.assertEqual(Artist.select().count(), 3)
        self.assertEqual(Album.select().count(), 6)
        self.assertEqual(Track.select().count(), 18)

    def test_get_music_folders(self):
        rv, child = self._make_request("getMusicFolders", tag="musicFolders")
        self.assertEqual(len(child), 2)
        self.assertSequenceEqual(
            sorted(self._xpath(child, "./musicFolder/@name")),
            ["Empty root", "Root folder"],
        )

    def test_get_indexes(self):
        self._make_request("getIndexes", {"musicFolderId": "abcdef"}, error=0)
        self._make_request("getIndexes", {"musicFolderId": 1234567890}, error=70)
        self._make_request("getIndexes", {"ifModifiedSince": "quoi"}, error=0)

        rv, child = self._make_request(
            "getIndexes",
            {"ifModifiedSince": int(time.time() * 1000 + 1000)},
            tag="indexes",
        )
        self.assertEqual(len(child), 0)

        fid = Folder.get(name="Empty root").id
        rv, child = self._make_request(
            "getIndexes", {"musicFolderId": str(fid)}, tag="indexes"
        )
        self.assertEqual(len(child), 0)

        rv, child = self._make_request("getIndexes", tag="indexes")
        self.assertEqual(len(child), 3)
        for i, letter in enumerate(["A", "B", "C"]):
            self.assertEqual(child[i].get("name"), letter)
            self.assertEqual(len(child[i]), 1)
            self.assertEqual(child[i][0].get("name"), letter + "rtist")

    def test_get_music_directory(self):
        self._make_request("getMusicDirectory", error=10)
        self._make_request("getMusicDirectory", {"id": "id"}, error=0)
        self._make_request("getMusicDirectory", {"id": 1234567890}, error=70)

        # should test with folders with both children folders and tracks. this code would break in that case
        for f in Folder.select():
            rv, child = self._make_request(
                "getMusicDirectory", {"id": str(f.id)}, tag="directory"
            )
            self.assertEqual(child.get("id"), str(f.id))
            self.assertEqual(child.get("name"), f.name)
            self.assertEqual(len(child), f.children.count() + f.tracks.count())
            for dbc, xmlc in zip(
                sorted(f.children, key=lambda c: c.name),
                sorted(child, key=lambda c: c.get("title")),
            ):
                self.assertEqual(dbc.name, xmlc.get("title"))
                self.assertEqual(xmlc.get("artist"), f.name)
                self.assertEqual(xmlc.get("parent"), str(f.id))
            for t, xmlc in zip(
                sorted(f.tracks, key=lambda t: t.title),
                sorted(child, key=lambda c: c.get("title")),
            ):
                self.assertEqual(t.title, xmlc.get("title"))
                self.assertEqual(xmlc.get("parent"), str(f.id))

    def test_get_artists(self):
        # same as getIndexes standard case
        # dataset should be improved to have a different directory structure than /root/Artist/Album/Track

        _, child = self._make_request("getArtists", tag="artists")
        self.assertEqual(len(child), 3)
        for i, letter in enumerate(["A", "B", "C"]):
            self.assertEqual(child[i].get("name"), letter)
            self.assertEqual(len(child[i]), 1)
            self.assertEqual(child[i][0].get("name"), letter + "rtist")

        self._make_request("getArtists", {"musicFolderId": "id"}, error=0)
        self._make_request("getArtists", {"musicFolderId": -3}, error=70)

        _, child = self._make_request(
            "getArtists", {"musicFolderId": str(self.empty_root.id)}, tag="artists"
        )
        self.assertEqual(len(child), 0)

        _, child = self._make_request(
            "getArtists", {"musicFolderId": str(self.root.id)}, tag="artists"
        )
        self.assertEqual(len(child), 3)

    def test_get_artist(self):
        # dataset should be improved to have tracks by a different artist than the album's artist
        self._make_request("getArtist", error=10)
        self._make_request("getArtist", {"id": "artist"}, error=0)
        self._make_request("getArtist", {"id": str(uuid.uuid4())}, error=70)

        for ar in Artist.select():
            rv, child = self._make_request(
                "getArtist", {"id": str(ar.id)}, tag="artist"
            )
            self.assertEqual(child.get("id"), str(ar.id))
            self.assertEqual(child.get("albumCount"), str(len(child)))
            self.assertEqual(len(child), ar.albums.count())
            for dal, xal in zip(
                sorted(ar.albums, key=lambda a: a.name),
                sorted(child, key=lambda c: c.get("name")),
            ):
                self.assertEqual(dal.name, xal.get("name"))
                self.assertEqual(
                    xal.get("artist"), ar.name
                )  # could break with a better dataset
                self.assertEqual(xal.get("artistId"), str(ar.id))  # see above

    def test_get_album(self):
        self._make_request("getAlbum", error=10)
        self._make_request("getAlbum", {"id": "nastynasty"}, error=0)
        self._make_request("getAlbum", {"id": str(uuid.uuid4())}, error=70)

        a = Album.select().first()
        rv, child = self._make_request("getAlbum", {"id": str(a.id)}, tag="album")
        self.assertEqual(child.get("id"), str(a.id))
        self.assertEqual(child.get("songCount"), str(len(child)))

        self.assertEqual(len(child), a.tracks.count())
        for dal, xal in zip(
            sorted(a.tracks, key=lambda t: t.title),
            sorted(child, key=lambda c: c.get("title")),
        ):
            self.assertEqual(dal.title, xal.get("title"))
            self.assertEqual(xal.get("album"), a.name)
            self.assertEqual(xal.get("albumId"), str(a.id))

    def test_get_song(self):
        self._make_request("getSong", error=10)
        self._make_request("getSong", {"id": "nastynasty"}, error=0)
        self._make_request("getSong", {"id": str(uuid.uuid4())}, error=70)

        s = Track.select().first()
        self._make_request("getSong", {"id": str(s.id)}, tag="song")

    def test_get_videos(self):
        self._make_request("getVideos", error=0)

    def test_genres(self):
        rv, child = self._make_request("getGenres", tag="genres")
        self.assertEqual(len(child), 1)
        self.assertEqual(child[0].text, "Music!")
        self.assertEqual(child[0].get("songCount"), "18")
        self.assertEqual(child[0].get("albumCount"), "6")


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017 Alban 'spl0k' Féron
#               2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest

from .apitestbase import ApiTestBase


class SystemTestCase(ApiTestBase):
    def test_ping(self):
        self._make_request("ping")

    def test_get_license(self):
        rv, child = self._make_request("getLicense", tag="license")
        self.assertEqual(child.get("valid"), "true")


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2018-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from . import api
from .exceptions import GenericError

methods = (
    "getVideos",
    "getAvatar",
    "getShares",
    "createShare",
    "updateShare",
    "deleteShare",
    "hls",
)


def unsupported():
    return GenericError("Not supported by Supysonic")


for m in methods:
    api.add_url_rule(f"/{m}", "unsupported", unsupported, methods=["GET", "POST"])
    api.add_url_rule(f"/{m}.view", "unsupported", unsupported, methods=["GET", "POST"])



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2020-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import os
import os.path
import shutil
import sys
import tempfile
import unittest

from supysonic.db import init_database, release_database
from supysonic.managers.folder import FolderManager
from supysonic.scanner import Scanner


@unittest.skipIf(
    sys.platform == "win32", "Windows doesn't seem too allow badly encoded paths"
)
class Issue85TestCase(unittest.TestCase):
    def setUp(self):
        self.__dir = tempfile.mkdtemp()
        init_database("sqlite:")
        FolderManager.add("folder", self.__dir)

    def tearDown(self):
        release_database()
        shutil.rmtree(self.__dir)

    def test_issue(self):
        os.mkdir(os.path.join(self.__dir.encode(), b"\xe6"))
        shutil.copyfile(
            "tests/assets/folder/silence.mp3",
            os.path.join(self.__dir.encode(), b"\xe6", b"silence.mp3"),
        )

        scanner = Scanner()
        scanner.queue_folder("folder")
        scanner.run()


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from collections import OrderedDict
from datetime import datetime
from flask import request

from ..db import Folder, Track, Artist, Album

from . import api_routing, get_root_folder
from .exceptions import MissingParameter


@api_routing("/search")
def old_search():
    artist, album, title, anyf, count, offset, newer_than = map(
        request.values.get,
        ("artist", "album", "title", "any", "count", "offset", "newerThan"),
    )

    count = int(count) if count else 20
    offset = int(offset) if offset else 0
    newer_than = int(newer_than) / 1000 if newer_than else 0
    min_date = datetime.fromtimestamp(newer_than)

    if artist:
        Child = Folder.alias()
        query = (
            Folder.select()
            .join(Child, on=Child.parent == Folder.id)
            .join(Track, on=Track.folder == Child.id)
            .where(Folder.name.contains(artist), Folder.created > min_date)
            .distinct()
        )
    elif album:
        query = (
            Folder.select()
            .join(Track, on=Track.folder)
            .where(Folder.name.contains(album), Folder.created > min_date)
            .distinct()
        )
    elif title:
        query = Track.select().where(
            Track.title.contains(title), Track.created > min_date
        )
    elif anyf:
        folders = Folder.select().where(
            Folder.name.contains(anyf), Folder.created > min_date
        )
        tracks = Track.select().where(
            Track.title.contains(anyf), Track.created > min_date
        )
        res = folders[offset : offset + count]
        fcount = folders.count()
        if offset + count > fcount:
            toff = max(0, offset - fcount)
            tend = offset + count - fcount
            res = res[:] + tracks[toff:tend][:]

        return request.formatter(
            "searchResult",
            {
                "totalHits": folders.count() + tracks.count(),
                "offset": offset,
                "match": [
                    (
                        r.as_subsonic_child(request.user)
                        if isinstance(r, Folder)
                        else r.as_subsonic_child(request.user, request.client)
                    )
                    for r in res
                ],
            },
        )
    else:
        raise MissingParameter("search")

    return request.formatter(
        "searchResult",
        {
            "totalHits": query.count(),
            "offset": offset,
            "match": [
                (
                    r.as_subsonic_child(request.user)
                    if isinstance(r, Folder)
                    else r.as_subsonic_child(request.user, request.client)
                )
                for r in query[offset : offset + count]
            ],
        },
    )


@api_routing("/search2")
def new_search():
    query = request.values["query"]
    (
        artist_count,
        artist_offset,
        album_count,
        album_offset,
        song_count,
        song_offset,
        mfid,
    ) = map(
        request.values.get,
        (
            "artistCount",
            "artistOffset",
            "albumCount",
            "albumOffset",
            "songCount",
            "songOffset",
            "musicFolderId",
        ),
    )

    artist_count = int(artist_count) if artist_count else 20
    artist_offset = int(artist_offset) if artist_offset else 0
    album_count = int(album_count) if album_count else 20
    album_offset = int(album_offset) if album_offset else 0
    song_count = int(song_count) if song_count else 20
    song_offset = int(song_offset) if song_offset else 0
    root = get_root_folder(mfid)

    Child = Folder.alias()
    artists = (
        Folder.select()
        .join(Child, on=Child.parent == Folder.id)
        .join(Track, on=Track.folder == Child.id)
        .where(Folder.name.contains(query))
        .distinct()
    )
    albums = (
        Folder.select()
        .join(Track, on=Track.folder)
        .where(Folder.name.contains(query))
        .distinct()
    )
    songs = Track.select().where(Track.title.contains(query))

    if root is not None:
        artists = artists.where(Track.root_folder == root)
        albums = albums.where(Track.root_folder == root)
        songs = songs.where(Track.root_folder == root)

    artists = artists.limit(artist_count).offset(artist_offset)
    albums = albums.limit(album_count).offset(album_offset)
    songs = songs.limit(song_count).offset(song_offset)

    return request.formatter(
        "searchResult2",
        OrderedDict(
            (
                ("artist", [a.as_subsonic_artist(request.user) for a in artists]),
                ("album", [f.as_subsonic_child(request.user) for f in albums]),
                (
                    "song",
                    [t.as_subsonic_child(request.user, request.client) for t in songs],
                ),
            )
        ),
    )


@api_routing("/search3")
def search_id3():
    query = request.values["query"]
    (
        artist_count,
        artist_offset,
        album_count,
        album_offset,
        song_count,
        song_offset,
        mfid,
    ) = map(
        request.values.get,
        (
            "artistCount",
            "artistOffset",
            "albumCount",
            "albumOffset",
            "songCount",
            "songOffset",
            "musicFolderId",
        ),
    )

    artist_count = int(artist_count) if artist_count else 20
    artist_offset = int(artist_offset) if artist_offset else 0
    album_count = int(album_count) if album_count else 20
    album_offset = int(album_offset) if album_offset else 0
    song_count = int(song_count) if song_count else 20
    song_offset = int(song_offset) if song_offset else 0
    root = get_root_folder(mfid)

    artists = Artist.select().where(Artist.name.contains(query))
    albums = Album.select().where(Album.name.contains(query))
    songs = Track.select().where(Track.title.contains(query))

    if root is not None:
        artists = artists.join(Track).where(Track.root_folder == root)
        albums = albums.join(Track).where(Track.root_folder == root)
        songs = songs.where(Track.root_folder == root)

    artists = artists.limit(artist_count).offset(artist_offset)
    albums = albums.limit(album_count).offset(album_offset)
    songs = songs.limit(song_count).offset(song_offset)

    return request.formatter(
        "searchResult3",
        OrderedDict(
            (
                ("artist", [a.as_subsonic_artist(request.user) for a in artists]),
                ("album", [a.as_subsonic_album(request.user) for a in albums]),
                (
                    "song",
                    [t.as_subsonic_child(request.user, request.client) for t in songs],
                ),
            )
        ),
    )



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import flask.json
import os.path
import requests
import unittest

from supysonic.db import Folder, Artist, Album, Track

from ..api.apitestbase import ApiTestBase


class LyricsTestCase(ApiTestBase):
    def setUp(self):
        super().setUp()

        self.config.WEBAPP["online_lyrics"] = True

        folder = Folder.create(
            name="Root",
            path=os.path.abspath("tests/assets/lyrics"),
            root=True,
        )

        artist = Artist.create(name="Artist")
        album = Album.create(artist=artist, name="Album")

        Track.create(
            title="Nope",
            number=1,
            disc=1,
            artist=artist,
            album=album,
            path=os.path.abspath("tests/assets/lyrics/empty.mp3"),
            root_folder=folder,
            folder=folder,
            duration=2,
            bitrate=320,
            last_modification=0,
        )
        Track.create(
            title="Yay",
            number=1,
            disc=1,
            artist=artist,
            album=album,
            path=os.path.abspath("tests/assets/lyrics/withlyrics.mp3"),
            root_folder=folder,
            folder=folder,
            duration=2,
            bitrate=320,
            last_modification=0,
        )

    def test_get_lyrics(self):
        self._make_request("getLyrics", error=10)
        self._make_request("getLyrics", {"artist": "artist"}, error=10)
        self._make_request("getLyrics", {"title": "title"}, error=10)

        # Potentially skip the tests if ChartLyrics is down (which happens quite often)
        try:
            requests.get("http://api.chartlyrics.com/", timeout=5)
        except requests.exceptions.Timeout:
            self.skipTest("ChartLyrics down")

        rv, child = self._make_request(
            "getLyrics",
            {
                "artist": "some really long name hoping",
                "title": "to get absolutely no result",
            },
            tag="lyrics",
        )
        self.assertIsNone(child.text)

        # ChartLyrics
        rv, child = self._make_request(
            "getLyrics",
            {"artist": "The Clash", "title": "London Calling"},
            tag="lyrics",
        )
        self.assertIn("live by the river", child.text)
        # ChartLyrics, JSON format
        args = {
            "u": "alice",
            "p": "Alic3",
            "c": "tests",
            "f": "json",
            "artist": "The Clash",
            "title": "London Calling",
        }
        rv = self.client.get("/rest/getLyrics.view", query_string=args)
        json = flask.json.loads(rv.data)
        self.assertIn("value", json["subsonic-response"]["lyrics"])
        self.assertIn("live by the river", json["subsonic-response"]["lyrics"]["value"])

        # Local file
        rv, child = self._make_request(
            "getLyrics", {"artist": "artist", "title": "nope"}, tag="lyrics"
        )
        self.assertIn("text file", child.text)

        # Metadata
        rv, child = self._make_request(
            "getLyrics", {"artist": "artist", "title": "yay"}, tag="lyrics"
        )
        self.assertIn("Some words", child.text)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2019 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from multiprocessing.connection import Client

from .exceptions import DaemonUnavailableError
from ..config import get_current_config
from ..utils import get_secret_key

__all__ = ["DaemonClient"]


class DaemonCommand:
    def apply(self, connection, daemon):
        raise NotImplementedError()


class WatcherCommand(DaemonCommand):
    def __init__(self, folder):
        self._folder = folder


class AddWatchedFolderCommand(WatcherCommand):
    def apply(self, connection, daemon):
        if daemon.watcher is not None:
            daemon.watcher.add_folder(self._folder)


class RemoveWatchedFolder(WatcherCommand):
    def apply(self, connection, daemon):
        if daemon.watcher is not None:
            daemon.watcher.remove_folder(self._folder)


class ScannerCommand(DaemonCommand):
    pass


class ScannerProgressCommand(ScannerCommand):
    def apply(self, connection, daemon):
        scanner = daemon.scanner
        rv = scanner.scanned if scanner is not None and scanner.is_alive() else None
        connection.send(ScannerProgressResult(rv))


class ScannerStartCommand(ScannerCommand):
    def __init__(self, folders=[], force=False):
        self.__folders = folders
        self.__force = force

    def apply(self, connection, daemon):
        daemon.start_scan(self.__folders, self.__force)


class JukeboxCommand(DaemonCommand):
    def __init__(self, action, args):
        self.__action = action
        self.__args = args

    def apply(self, connection, daemon):
        if daemon.jukebox is None:
            connection.send(JukeboxResult(None))
            return

        playlist = None
        if self.__action == "get":
            playlist = daemon.jukebox.playlist
        elif self.__action == "status":
            pass
        else:
            func = None

            if self.__action == "set":
                func = daemon.jukebox.set
            elif self.__action == "start":
                func = daemon.jukebox.start
            elif self.__action == "stop":
                func = daemon.jukebox.stop
            elif self.__action == "skip":
                func = daemon.jukebox.skip
            elif self.__action == "add":
                func = daemon.jukebox.add
            elif self.__action == "clear":
                func = daemon.jukebox.clear
            elif self.__action == "remove":
                func = daemon.jukebox.remove
            elif self.__action == "shuffle":
                func = daemon.jukebox.shuffle
            elif self.__action == "setGain":
                func = daemon.jukebox.setgain

            func(*self.__args)

        rv = JukeboxResult(daemon.jukebox)
        rv.playlist = playlist
        connection.send(rv)


class DaemonCommandResult:
    pass


class ScannerProgressResult(DaemonCommandResult):
    def __init__(self, scanned):
        self.__scanned = scanned

    scanned = property(lambda self: self.__scanned)


class JukeboxResult(DaemonCommandResult):
    def __init__(self, jukebox):
        if jukebox is None:
            self.playing = False
            self.index = -1
            self.gain = 1.0
            self.position = 0
        else:
            self.playing = jukebox.playing
            self.index = jukebox.index
            self.gain = jukebox.gain
            self.position = jukebox.position
        self.playlist = ()


class DaemonClient:
    def __init__(self, address=None):
        self.__address = address or get_current_config().DAEMON["socket"]
        self.__key = get_secret_key("daemon_key")

    def __get_connection(self):
        if not self.__address:
            raise DaemonUnavailableError("No daemon address set")
        try:
            return Client(address=self.__address, authkey=self.__key)
        except OSError:
            raise DaemonUnavailableError(
                f"Couldn't connect to daemon at {self.__address}"
            )

    def add_watched_folder(self, folder):
        if not isinstance(folder, str):
            raise TypeError("Expecting string, got " + str(type(folder)))
        with self.__get_connection() as c:
            c.send(AddWatchedFolderCommand(folder))

    def remove_watched_folder(self, folder):
        if not isinstance(folder, str):
            raise TypeError("Expecting string, got " + str(type(folder)))
        with self.__get_connection() as c:
            c.send(RemoveWatchedFolder(folder))

    def get_scanning_progress(self):
        with self.__get_connection() as c:
            c.send(ScannerProgressCommand())
            return c.recv().scanned

    def scan(self, folders=[], force=False):
        if not isinstance(folders, (list, tuple)):
            raise TypeError("Expecting list, got " + str(type(folders)))
        with self.__get_connection() as c:
            c.send(ScannerStartCommand(folders, force))

    def jukebox_control(self, action, *args):
        if not isinstance(action, str):
            raise TypeError("Expecting string, got " + str(type(action)))
        with self.__get_connection() as c:
            c.send(JukeboxCommand(action, args))
            return c.recv()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

from flask import (
    current_app,
    flash,
    redirect,
    request,
    render_template,
    session,
    url_for,
)
from flask import Blueprint
from functools import wraps

from .. import VERSION, DOWNLOAD_URL
from ..daemon.client import DaemonClient
from ..daemon.exceptions import DaemonUnavailableError
from ..db import Artist, Album, Track
from ..managers.user import UserManager

frontend = Blueprint("frontend", __name__)


@frontend.context_processor
def inject_metadata():
    return {"version": VERSION, "download_url": DOWNLOAD_URL}


@frontend.before_request
def login_check():
    request.user = None
    should_login = True
    if session.get("userid"):
        try:
            user = UserManager.get(session.get("userid"))
            request.user = user
            should_login = False
        except (ValueError, User.DoesNotExist):
            session.clear()

    if should_login and request.endpoint != "frontend.login":
        flash("Please login")
        return redirect(
            url_for(
                "frontend.login",
                returnUrl=request.script_root
                + request.url[len(request.url_root) - 1 :],
            )
        )


@frontend.before_request
def scan_status():
    if not request.user or not request.user.admin:
        return

    try:
        scanned = DaemonClient(
            current_app.config["DAEMON"]["socket"]
        ).get_scanning_progress()
        if scanned is not None:
            flash(f"Scanning in progress, {scanned} files scanned.")
    except DaemonUnavailableError:
        pass


@frontend.route("/")
def index():
    stats = {
        "artists": Artist.select().count(),
        "albums": Album.select().count(),
        "tracks": Track.select().count(),
    }
    return render_template("home.html", stats=stats)


def admin_only(f):
    @wraps(f)
    def decorated_func(*args, **kwargs):
        if not request.user or not request.user.admin:
            return redirect(url_for("frontend.index"))
        return f(*args, **kwargs)

    return decorated_func


from .user import *
from .folder import *
from .playlist import *



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest
import flask.json

from xml.etree import ElementTree

from supysonic.api.formatters import JSONFormatter, JSONPFormatter, XMLFormatter

from ..testbase import TestBase


class UnwrapperMixin:
    def make_response(self, elem, data):
        with self.request_context():
            rv = super().make_response(elem, data)
            return rv.get_data(as_text=True)

    @staticmethod
    def create_from(cls):
        class Unwrapper(UnwrapperMixin, cls):
            pass

        return Unwrapper


class ResponseHelperJsonTestCase(TestBase, UnwrapperMixin.create_from(JSONFormatter)):
    def make_response(self, elem, data):
        rv = super().make_response(elem, data)
        return flask.json.loads(rv)

    def process_and_extract(self, d):
        return self.make_response("tag", d)["subsonic-response"]["tag"]

    def test_basic(self):
        empty = self.empty
        self.assertEqual(len(empty), 1)
        self.assertIn("subsonic-response", empty)
        self.assertIsInstance(empty["subsonic-response"], dict)

        resp = empty["subsonic-response"]
        self.assertEqual(len(resp), 2)
        self.assertIn("status", resp)
        self.assertIn("version", resp)
        self.assertEqual(resp["status"], "ok")

        resp = self.error(0, "message")["subsonic-response"]
        self.assertEqual(resp["status"], "failed")

        some_dict = {"intValue": 2, "someString": "Hello world!"}
        resp = self.process_and_extract(some_dict)
        self.assertIn("intValue", resp)
        self.assertIn("someString", resp)

    def test_lists(self):
        resp = self.process_and_extract({"someList": [2, 4, 8, 16], "emptyList": []})
        self.assertIn("someList", resp)
        self.assertNotIn("emptyList", resp)
        self.assertListEqual(resp["someList"], [2, 4, 8, 16])

    def test_dicts(self):
        resp = self.process_and_extract({"dict": {"s": "Blah", "i": 20}, "empty": {}})
        self.assertIn("dict", resp)
        self.assertIn("empty", resp)
        self.assertDictEqual(resp["dict"], {"s": "Blah", "i": 20})
        self.assertDictEqual(resp["empty"], {})

    def test_nesting(self):
        resp = self.process_and_extract(
            {
                "dict": {
                    "value": "hey look! a string",
                    "list": [1, 2, 3],
                    "emptyList": [],
                    "subdict": {"a": "A"},
                },
                "list": [{"b": "B"}, {"c": "C"}, [4, 5, 6], "final string"],
            }
        )

        self.assertEqual(len(resp), 2)
        self.assertIn("dict", resp)
        self.assertIn("list", resp)

        dct = resp["dict"]
        lst = resp["list"]

        self.assertIn("value", dct)
        self.assertIn("list", dct)
        self.assertNotIn("emptyList", dct)
        self.assertIn("subdict", dct)
        self.assertIsInstance(dct["value"], str)
        self.assertIsInstance(dct["list"], list)
        self.assertIsInstance(dct["subdict"], dict)

        self.assertEqual(lst, [{"b": "B"}, {"c": "C"}, [4, 5, 6], "final string"])


class ResponseHelperJsonpTestCase(TestBase, UnwrapperMixin.create_from(JSONPFormatter)):
    def test_basic(self):
        self._JSONPFormatter__callback = "callback"  # hacky
        result = self.empty
        self.assertTrue(result.startswith("callback({"))
        self.assertTrue(result.endswith("})"))

        json = flask.json.loads(result[9:-1])
        self.assertIn("subsonic-response", json)


class ResponseHelperXMLTestCase(TestBase, UnwrapperMixin.create_from(XMLFormatter)):
    def make_response(self, elem, data):
        xml = super().make_response(elem, data)
        xml = xml.replace('xmlns="http://subsonic.org/restapi"', "")
        root = ElementTree.fromstring(xml)
        return root

    def process_and_extract(self, d):
        rv = self.make_response("tag", d)
        return rv.find("tag")

    def assertAttributesMatchDict(self, elem, d):
        d = {k: str(v) for k, v in d.items()}
        self.assertDictEqual(elem.attrib, d)

    def test_root(self):
        xml = super().make_response("tag", {})
        self.assertIn("<subsonic-response ", xml)
        self.assertIn('xmlns="http://subsonic.org/restapi"', xml)
        self.assertTrue(xml.strip().endswith("</subsonic-response>"))

    def test_basic(self):
        empty = self.empty
        self.assertIsNotNone(empty.find(".[@version]"))
        self.assertIsNotNone(empty.find(".[@status='ok']"))

        resp = self.error(0, "message")
        self.assertIsNotNone(resp.find(".[@status='failed']"))

        some_dict = {"intValue": 2, "someString": "Hello world!"}
        resp = self.process_and_extract(some_dict)
        self.assertIsNotNone(resp.find(".[@intValue]"))
        self.assertIsNotNone(resp.find(".[@someString]"))

    def test_lists(self):
        resp = self.process_and_extract({"someList": [2, 4, 8, 16], "emptyList": []})

        elems = resp.findall("./someList")
        self.assertEqual(len(elems), 4)
        self.assertIsNone(resp.find("./emptyList"))

        for e, i in zip(elems, [2, 4, 8, 16]):
            self.assertEqual(int(e.text), i)

    def test_dicts(self):
        resp = self.process_and_extract({"dict": {"s": "Blah", "i": 20}, "empty": {}})

        d = resp.find("./dict")
        self.assertIsNotNone(d)
        self.assertIsNotNone(resp.find("./empty"))
        self.assertAttributesMatchDict(d, {"s": "Blah", "i": 20})

    def test_nesting(self):
        resp = self.process_and_extract(
            {
                "dict": {
                    "somevalue": "hey look! a string",
                    "list": [1, 2, 3],
                    "emptyList": [],
                    "subdict": {"a": "A"},
                },
                "list": [{"b": "B"}, {"c": "C"}, "final string"],
            }
        )

        self.assertEqual(len(resp), 4)  # 'dict' and 3 'list's

        d = resp.find("./dict")
        lists = resp.findall("./list")

        self.assertIsNotNone(d)
        self.assertAttributesMatchDict(d, {"somevalue": "hey look! a string"})
        self.assertEqual(len(d.findall("./list")), 3)
        self.assertEqual(len(d.findall("./emptyList")), 0)
        self.assertIsNotNone(d.find("./subdict"))

        self.assertEqual(len(lists), 3)
        self.assertAttributesMatchDict(lists[0], {"b": "B"})
        self.assertAttributesMatchDict(lists[1], {"c": "C"})
        self.assertEqual(lists[2].text, "final string")


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2019-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import uuid

from flask import current_app, request

from ..daemon import DaemonClient
from ..daemon.exceptions import DaemonUnavailableError
from ..db import Track

from . import api_routing
from .exceptions import GenericError, MissingParameter, Forbidden


@api_routing("/jukeboxControl")
def jukebox_control():
    if not request.user.jukebox and not request.user.admin:
        raise Forbidden()

    action = request.values["action"]

    index = request.values.get("index")
    offset = request.values.get("offset")
    id = request.values.getlist("id")
    gain = request.values.get("gain")

    if action not in (
        "get",
        "status",
        "set",
        "start",
        "stop",
        "skip",
        "add",
        "clear",
        "remove",
        "shuffle",
        "setGain",
    ):
        raise GenericError("Unknown action")

    args = ()
    if action == "set":
        if id:
            args = [uuid.UUID(i) for i in id]
    elif action == "skip":
        if not index:
            raise MissingParameter("index")
        if offset:
            args = (int(index), int(offset))
        else:
            args = (int(index), 0)
    elif action == "add":
        if not id:
            raise MissingParameter("id")
        else:
            args = [uuid.UUID(i) for i in id]
    elif action == "remove":
        if not index:
            raise MissingParameter("index")
        else:
            args = (int(index),)
    elif action == "setGain":
        if not gain:
            raise MissingParameter("gain")
        else:
            args = (float(gain),)

    try:
        status = DaemonClient(current_app.config["DAEMON"]["socket"]).jukebox_control(
            action, *args
        )
    except DaemonUnavailableError:
        raise GenericError("Jukebox unavaliable")

    rv = {
        "currentIndex": status.index,
        "playing": status.playing,
        "gain": status.gain,
        "position": status.position,
    }
    if action == "get":
        playlist = []
        for path in status.playlist:
            try:
                playlist.append(Track.get(path=path))
            except Track.DoesNotExist:
                pass
        rv["entry"] = [
            t.as_subsonic_child(request.user, request.client) for t in playlist
        ]
        return request.formatter("jukeboxPlaylist", rv)
    else:
        return request.formatter("jukeboxStatus", rv)



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2021 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from . import main

if __name__ == "__main__":
    main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from flask import current_app, flash, redirect, render_template, request, url_for

from ..daemon.client import DaemonClient
from ..daemon.exceptions import DaemonUnavailableError
from ..db import Folder
from ..managers.folder import FolderManager

from . import admin_only, frontend


@frontend.route("/folder")
@admin_only
def folder_index():
    try:
        DaemonClient(current_app.config["DAEMON"]["socket"]).get_scanning_progress()
        allow_scan = True
    except DaemonUnavailableError:
        allow_scan = False
        flash(
            "The daemon is unavailable, can't scan from the web interface, use the CLI to do so.",
            "warning",
        )
    return render_template(
        "folders.html",
        folders=Folder.select().where(Folder.root),
        allow_scan=allow_scan,
    )


@frontend.route("/folder/add")
@admin_only
def add_folder_form():
    return render_template("addfolder.html")


@frontend.route("/folder/add", methods=["POST"])
@admin_only
def add_folder_post():
    error = False
    name, path = map(request.form.get, ("name", "path"))
    if name in (None, ""):
        flash("The name is required.")
        error = True
    if path in (None, ""):
        flash("The path is required.")
        error = True
    if error:
        return render_template("addfolder.html")

    try:
        FolderManager.add(name, path)
    except ValueError as e:
        flash(str(e), "error")
        return render_template("addfolder.html")

    flash(f"Folder '{name}' created. You should now run a scan")
    return redirect(url_for("frontend.folder_index"))


@frontend.route("/folder/del/<id>")
@admin_only
def del_folder(id):
    try:
        FolderManager.delete(id)
        flash("Deleted folder")
    except ValueError as e:
        flash(str(e), "error")
    except Folder.DoesNotExist:
        flash("No such folder", "error")

    return redirect(url_for("frontend.folder_index"))


@frontend.route("/folder/scan")
@frontend.route("/folder/scan/<id>")
@admin_only
def scan_folder(id=None):
    try:
        if id is not None:
            folders = [FolderManager.get(id).name]
        else:
            folders = []
        DaemonClient(current_app.config["DAEMON"]["socket"]).scan(folders)
        flash("Scanning started")
    except ValueError as e:
        flash(str(e), "error")
    except Folder.DoesNotExist:
        flash("No such folder", "error")
    except DaemonUnavailableError:
        flash("Can't start scan", "error")

    return redirect(url_for("frontend.folder_index"))



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2019-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from base64 import b64encode, b64decode
from os import urandom

from supysonic.db import Meta


__key_cache = {}


def get_secret_key(keyname):
    if keyname in __key_cache:
        return __key_cache[keyname]

    try:
        key = b64decode(Meta[keyname].value)
    except Meta.DoesNotExist:
        key = urandom(128)
        Meta.create(key=keyname, value=b64encode(key).decode())

    __key_cache[keyname] = key
    return key



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

API_VERSION = "1.12.0"

import binascii
import logging
import uuid
from flask import request
from flask import Blueprint
from peewee import IntegrityError

from ..db import ClientPrefs, Folder
from ..managers.user import UserManager

from .exceptions import GenericError, Unauthorized, NotFound
from .formatters import JSONFormatter, JSONPFormatter, XMLFormatter

api = Blueprint("api", __name__)
logger = logging.getLogger(__name__)


def api_routing(endpoint):
    def decorator(func):
        viewendpoint = f"{endpoint}.view"
        api.add_url_rule(endpoint, view_func=func, methods=["GET", "POST"])
        api.add_url_rule(viewendpoint, view_func=func, methods=["GET", "POST"])
        return func

    return decorator


@api.before_request
def set_formatter():
    """Return a function to create the response."""
    f, callback = map(request.values.get, ("f", "callback"))
    if f == "jsonp":
        request.formatter = JSONPFormatter(callback)
    elif f == "json":
        request.formatter = JSONFormatter()
    else:
        request.formatter = XMLFormatter()


def decode_password(password):
    if not password.startswith("enc:"):
        return password

    try:
        return binascii.unhexlify(password[4:].encode("utf-8")).decode("utf-8")
    except ValueError:
        return password


@api.before_request
def authorize():
    if request.authorization:
        username = request.authorization.username
        user = UserManager.try_auth(username, request.authorization.password)
        if user is not None:
            request.user = user
            return

        logger.error(
            "Failed login attempt for user %s (IP: %s)", username, request.remote_addr
        )
        raise Unauthorized()

    username = request.values["u"]
    password = request.values["p"]
    password = decode_password(password)

    user = UserManager.try_auth(username, password)
    if user is None:
        logger.error(
            "Failed login attempt for user %s (IP: %s)", username, request.remote_addr
        )
        raise Unauthorized()

    request.user = user


@api.before_request
def get_client_prefs():
    client = request.values["c"]
    try:
        request.client = ClientPrefs[request.user, client]
    except ClientPrefs.DoesNotExist:
        try:
            request.client = ClientPrefs.create(user=request.user, client_name=client)
        except IntegrityError:
            # We might have hit a race condition here, another request already created
            # the ClientPrefs. Issue #220
            request.client = ClientPrefs[request.user, client]


def get_entity(cls, param="id"):
    eid = request.values[param]
    if cls == Folder:
        eid = int(eid)
    else:
        eid = uuid.UUID(eid)
    return cls[eid]


def get_entity_id(cls, eid):
    """Return the entity ID as its proper type."""
    if cls == Folder:
        if isinstance(eid, uuid.UUID):
            raise GenericError("Invalid ID")
        try:
            return int(eid)
        except ValueError as e:
            raise GenericError("Invalid ID") from e
    try:
        return uuid.UUID(eid)
    except (AttributeError, ValueError) as e:
        raise GenericError("Invalid ID") from e


def get_root_folder(id):
    if id is None:
        return None

    try:
        fid = int(id)
    except ValueError as e:
        raise ValueError("Invalid folder ID") from e

    try:
        return Folder.get(id=fid, root=True)
    except Folder.DoesNotExist as e:
        raise NotFound("Folder") from e


from .errors import *

from .system import *
from .browse import *
from .user import *
from .albums_songs import *
from .media import *
from .annotation import *
from .chat import *
from .search import *
from .playlists import *
from .jukebox import *
from .radio import *
from .unsupported import *
from .scan import *



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2018-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from peewee import DoesNotExist
from werkzeug.exceptions import BadRequestKeyError

from . import api
from .exceptions import GenericError, MissingParameter, NotFound, ServerError


@api.errorhandler(ValueError)
def value_error(e):
    return GenericError("{0.__class__.__name__}: {0}".format(e))


@api.errorhandler(BadRequestKeyError)
def key_error(e):
    return MissingParameter()


@api.errorhandler(DoesNotExist)
def object_not_found(e):
    return NotFound(e.__class__.__name__[: -len("DoesNotExist")])


@api.errorhandler(500)
def generic_error(e):  # pragma: nocover
    return ServerError("{0.__class__.__name__}: {0}".format(e))


# @api.errorhandler(404)
@api.route("/<path:invalid>", methods=["GET", "POST"])  # blueprint 404 workaround
def not_found(*args, **kwargs):
    return GenericError("Unknown method"), 404



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2014-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import logging

from logging.handlers import TimedRotatingFileHandler
from signal import signal, SIGTERM, SIGINT

from .client import DaemonClient
from .server import Daemon

from ..config import IniConfig
from ..db import init_database, release_database

__all__ = ["Daemon", "DaemonClient"]

logger = logging.getLogger("supysonic")

daemon = None


def setup_logging(config):
    if config["log_file"]:
        if config["log_rotate"]:
            log_handler = TimedRotatingFileHandler(config["log_file"], when="midnight")
        else:
            log_handler = logging.FileHandler(config["log_file"])
        log_handler.setFormatter(
            logging.Formatter("%(asctime)s [%(levelname)s] %(message)s")
        )
    else:
        log_handler = logging.StreamHandler()
        log_handler.setFormatter(logging.Formatter("[%(levelname)s] %(message)s"))
    logger.addHandler(log_handler)
    if "log_level" in config:
        level = getattr(logging, config["log_level"].upper(), logging.NOTSET)
        logger.setLevel(level)


def __terminate(signum, frame):
    global daemon

    logger.debug("Got signal %i. Stopping...", signum)
    daemon.terminate()
    release_database()


def main():
    global daemon

    config = IniConfig.from_common_locations()
    setup_logging(config.DAEMON)

    signal(SIGTERM, __terminate)
    signal(SIGINT, __terminate)

    init_database(config.BASE["database_uri"])
    daemon = Daemon(config)
    daemon.run()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2021-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import importlib
import os
import os.path

from click import command, option, Option
from click.exceptions import UsageError, ClickException
from click.types import Choice


_servers = [
    e.name[:-3]
    for e in os.scandir(os.path.dirname(__file__))
    if not e.name.startswith("_") and e.name.endswith(".py")
]


class MutuallyExclusiveOption(Option):
    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        help = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(self.mutually_exclusive)
            kwargs["help"] = (
                "{}  NOTE: This argument is mutually exclusive with arguments: [{}].".format(
                    help, ex_str
                )
            )
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise UsageError(
                "Illegal usage: `{}` is mutually exclusive with arguments `{}`.".format(
                    self.name, ", ".join(self.mutually_exclusive)
                )
            )

        return super().handle_parse_result(ctx, opts, args)


def get_server(name):
    return importlib.import_module("." + name, __package__).server


def find_first_available_server():
    for module in _servers:
        try:
            return get_server(module)
        except ImportError:
            pass

    return None


@command()
@option(
    "-S",
    "--server",
    type=Choice(_servers),
    help="Specify which WSGI server to use. Pick the first available if none is set",
)
@option(
    "-h",
    "--host",
    default="0.0.0.0",
    show_default=True,
    help="Hostname or IP address on which to listen",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=("socket",),
)
@option(
    "-p",
    "--port",
    default=5722,
    show_default=True,
    help="TCP port on which to listen",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=("socket",),
)
@option(
    "-s",
    "--socket",
    help="Unix socket on which to bind to, Can't be used with --host and --port",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=("host", "port"),
)
@option(
    "--processes",
    type=int,
    help="Number of processes to spawn. May not be supported by all servers",
)
@option(
    "--threads",
    type=int,
    help="Number of threads used to process application logic. May not be supported by all servers",
)
def main(server, host, port, socket, processes, threads):
    """Starts the Supysonic web server"""

    if server is None:
        server = find_first_available_server()
        if server is None:
            raise ClickException(
                f"Couldn't load any server, please install one of {_servers}"
            )
    else:
        try:
            server = get_server(server)
        except ImportError:
            raise ClickException(f"Couldn't load {server}, please install it first")

    if socket is not None:
        host = None
        port = None

    server(
        host=host, port=port, socket=socket, processes=processes, threads=threads
    ).run()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest
import sys

from flask import current_app

from supysonic.db import Track
from supysonic.managers.folder import FolderManager
from supysonic.scanner import Scanner

from .apitestbase import ApiTestBase


class TranscodingTestCase(ApiTestBase):
    def setUp(self):
        super().setUp()

        FolderManager.add("Folder", "tests/assets/folder")
        scanner = Scanner()
        scanner.queue_folder("Folder")
        scanner.run()

        self.trackid = Track.get().id

    def _stream(self, **kwargs):
        kwargs.update(
            {"u": "alice", "p": "Alic3", "c": "tests", "v": "1.9.0", "id": self.trackid}
        )

        rv = self.client.get("/rest/stream.view", query_string=kwargs)
        self.assertEqual(rv.status_code, 200)
        self.assertFalse(rv.mimetype.startswith("text/"))

        return rv

    def test_no_transcoding_available(self):
        self._make_request("stream", {"id": self.trackid, "format": "wat"}, error=0)

    @unittest.skipIf(
        sys.platform == "win32",
        "Can't test transcoding on Windows because of a lack of simple commandline tools",
    )
    def test_direct_transcode(self):
        rv = self._stream(maxBitRate=96, estimateContentLength="true")
        self.assertIn(b"tests/assets/folder/silence.mp3", rv.data)
        self.assertTrue(rv.data.endswith(b"96"))
        self.assertIn("Content-Length", rv.headers)
        self.assertEqual(rv.content_length, 48000)  # 4s at 96kbps

    @unittest.skipIf(
        sys.platform == "win32",
        "Can't test transcoding on Windows because of a lack of simple commandline tools",
    )
    def test_decode_encode(self):
        rv = self._stream(format="cat")
        self.assertEqual(rv.data, b"Pushing out some mp3 data...")

        rv = self._stream(format="md5")
        self.assertTrue(rv.data.startswith(b"dbb16c0847e5d8c3b1867604828cb50b"))

    @unittest.skipIf(
        sys.platform == "win32",
        "Can't test transcoding on Windows because of a lack of simple commandline tools",
    )
    def test_mostly_transcoded_cached(self):
        # See https://github.com/spl0k/supysonic/issues/202

        rv = self._stream(maxBitRate=96, estimateContentLength="true", format="rnd")

        read = 0
        it = iter(rv.response)
        # Read up to the estimated length
        while read < 48000:
            read += len(next(it))
        rv.response.close()
        rv.close()

        key = f"{self.trackid}-96.rnd"
        with self.app_context():
            self.assertTrue(current_app.transcode_cache.has(key))
            self.assertEqual(current_app.transcode_cache.size, 52000)

    @unittest.skipIf(
        sys.platform == "win32",
        "Can't test transcoding on Windows because of a lack of simple commandline tools",
    )
    def test_partly_transcoded_cached(self):
        rv = self._stream(maxBitRate=96, estimateContentLength="true", format="rnd")

        # read one check of data then close the connection
        next(iter(rv.response))
        rv.response.close()
        rv.close()

        key = f"{self.trackid}-96.rnd"
        with self.app_context():
            self.assertFalse(current_app.transcode_cache.has(key))
            self.assertEqual(current_app.transcode_cache.size, 0)

    @unittest.skipIf(
        sys.platform == "win32",
        "Can't test transcoding on Windows because of a lack of simple commandline tools",
    )
    def test_last_chunk_close_transcoded_cached(self):
        rv = self._stream(maxBitRate=96, estimateContentLength="true", format="rnd")

        read = 0
        it = iter(rv.response)
        # Read up to the last chunk of data but keep the generator "alive"
        while read < 52000:
            read += len(next(it))
        rv.response.close()
        rv.close()

        key = f"{self.trackid}-96.rnd"
        with self.app_context():
            self.assertTrue(current_app.transcode_cache.has(key))
            self.assertEqual(current_app.transcode_cache.size, 52000)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2020 Alban 'spl0k' Féron
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest

from ..utils import hexlify
from .apitestbase import ApiTestBase


class UserTestCase(ApiTestBase):
    def test_get_user(self):
        # missing username
        self._make_request("getUser", error=10)

        # non-existent user
        self._make_request("getUser", {"username": "non existent"}, error=70)

        # self
        rv, child = self._make_request("getUser", {"username": "alice"}, tag="user")
        self.assertEqual(child.get("username"), "alice")
        self.assertEqual(child.get("adminRole"), "true")

        # other
        rv, child = self._make_request("getUser", {"username": "bob"}, tag="user")
        self.assertEqual(child.get("username"), "bob")
        self.assertEqual(child.get("adminRole"), "false")

        # self from non-admin
        rv, child = self._make_request(
            "getUser", {"u": "bob", "p": "B0b", "username": "bob"}, tag="user"
        )
        self.assertEqual(child.get("username"), "bob")
        self.assertEqual(child.get("adminRole"), "false")

        # other from non-admin
        self._make_request(
            "getUser", {"u": "bob", "p": "B0b", "username": "alice"}, error=50
        )

    def test_get_users(self):
        # non-admin
        self._make_request("getUsers", {"u": "bob", "p": "B0b"}, error=50)

        # admin
        rv, child = self._make_request("getUsers", tag="users")
        self.assertEqual(len(child), 2)
        self.assertIsNotNone(self._find(child, "./user[@username='alice']"))
        self.assertIsNotNone(self._find(child, "./user[@username='bob']"))

    def test_create_user(self):
        # non admin
        self._make_request("createUser", {"u": "bob", "p": "B0b"}, error=50)

        # missing params, testing every combination, maybe overkill
        self._make_request("createUser", error=10)
        self._make_request("createUser", {"username": "user"}, error=10)
        self._make_request("createUser", {"password": "pass"}, error=10)
        self._make_request("createUser", {"email": "email@example.com"}, error=10)
        self._make_request(
            "createUser", {"username": "user", "password": "pass"}, error=10
        )
        self._make_request(
            "createUser", {"username": "user", "email": "email@example.com"}, error=10
        )
        self._make_request(
            "createUser", {"password": "pass", "email": "email@example.com"}, error=10
        )

        # duplicate
        self._make_request(
            "createUser",
            {"username": "bob", "password": "pass", "email": "me@bob.com"},
            error=0,
        )

        # test we only got our two initial users
        rv, child = self._make_request("getUsers", tag="users")
        self.assertEqual(len(child), 2)

        # create users
        self._make_request(
            "createUser",
            {
                "username": "charlie",
                "password": "Ch4rl1e",
                "email": "unicorn@example.com",
                "adminRole": True,
            },
            skip_post=True,
        )
        rv, child = self._make_request("getUser", {"username": "charlie"}, tag="user")
        self.assertEqual(child.get("username"), "charlie")
        self.assertEqual(child.get("email"), "unicorn@example.com")
        self.assertEqual(child.get("adminRole"), "true")
        self.assertEqual(child.get("jukeboxRole"), "true")  # admin gives full control

        self._make_request(
            "createUser",
            {
                "username": "dave",
                "password": "Dav3",
                "email": "dave@example.com",
                "jukeboxRole": True,
            },
            skip_post=True,
        )
        rv, child = self._make_request("getUser", {"username": "dave"}, tag="user")
        self.assertEqual(child.get("username"), "dave")
        self.assertEqual(child.get("email"), "dave@example.com")
        self.assertEqual(child.get("adminRole"), "false")
        self.assertEqual(child.get("jukeboxRole"), "true")

        self._make_request(
            "createUser",
            {"username": "eve", "password": "3ve", "email": "eve@example.com"},
            skip_post=True,
        )
        rv, child = self._make_request("getUser", {"username": "eve"}, tag="user")
        self.assertEqual(child.get("username"), "eve")
        self.assertEqual(child.get("email"), "eve@example.com")
        self.assertEqual(child.get("adminRole"), "false")
        self.assertEqual(child.get("jukeboxRole"), "false")

        rv, child = self._make_request("getUsers", tag="users")
        self.assertEqual(len(child), 5)

    def test_delete_user(self):
        # non admin
        self._make_request(
            "deleteUser", {"u": "bob", "p": "B0b", "username": "alice"}, error=50
        )

        # missing param
        self._make_request("deleteUser", error=10)

        # non existing
        self._make_request("deleteUser", {"username": "charlie"}, error=70)

        # test we still got our two initial users
        rv, child = self._make_request("getUsers", tag="users")
        self.assertEqual(len(child), 2)

        # delete user
        self._make_request("deleteUser", {"username": "bob"}, skip_post=True)
        rv, child = self._make_request("getUsers", tag="users")
        self.assertEqual(len(child), 1)

    def test_change_password(self):
        # missing parameter
        self._make_request("changePassword", error=10)
        self._make_request("changePassword", {"username": "alice"}, error=10)
        self._make_request("changePassword", {"password": "newpass"}, error=10)

        # admin change self
        self._make_request(
            "changePassword",
            {"username": "alice", "password": "newpass"},
            skip_post=True,
        )
        self._make_request("ping", error=40)
        self._make_request("ping", {"u": "alice", "p": "newpass"})
        self._make_request(
            "changePassword",
            {"u": "alice", "p": "newpass", "username": "alice", "password": "Alic3"},
            skip_post=True,
        )

        # admin change other
        self._make_request(
            "changePassword", {"username": "bob", "password": "newbob"}, skip_post=True
        )
        self._make_request("ping", {"u": "bob", "p": "B0b"}, error=40)
        self._make_request("ping", {"u": "bob", "p": "newbob"})

        # non-admin change self
        self._make_request(
            "changePassword",
            {"u": "bob", "p": "newbob", "username": "bob", "password": "B0b"},
            skip_post=True,
        )
        self._make_request("ping", {"u": "bob", "p": "newbob"}, error=40)
        self._make_request("ping", {"u": "bob", "p": "B0b"})

        # non-admin change other
        self._make_request(
            "changePassword",
            {"u": "bob", "p": "B0b", "username": "alice", "password": "newpass"},
            skip_post=True,
            error=50,
        )
        self._make_request("ping", {"u": "alice", "p": "newpass"}, error=40)
        self._make_request("ping")

        # change non existing
        self._make_request(
            "changePassword", {"username": "nonexsistent", "password": "pass"}, error=70
        )

        # non ASCII chars
        self._make_request(
            "changePassword",
            {"username": "alice", "password": "новыйпароль"},
            skip_post=True,
        )
        self._make_request("ping", {"u": "alice", "p": "новыйпароль"})
        self._make_request(
            "changePassword",
            {
                "username": "alice",
                "password": "Alic3",
                "u": "alice",
                "p": "новыйпароль",
            },
            skip_post=True,
        )

        # non ASCII in hex encoded password
        self._make_request(
            "changePassword",
            {"username": "alice", "password": "enc:" + hexlify("новыйпароль")},
            skip_post=True,
        )
        self._make_request("ping", {"u": "alice", "p": "новыйпароль"})

        # new password starting with 'enc:' followed by non hex chars
        self._make_request(
            "changePassword",
            {
                "username": "alice",
                "password": "enc:randomstring",
                "u": "alice",
                "p": "новыйпароль",
            },
            skip_post=True,
        )
        self._make_request("ping", {"u": "alice", "p": "enc:randomstring"})

    def test_update_user(self):
        # non admin
        self._make_request(
            "updateUser", {"u": "bob", "p": "B0b", "username": "alice"}, error=50
        )

        # missing param
        self._make_request("updateUser", error=10)

        # non existing
        self._make_request("updateUser", {"username": "charlie"}, error=70)

        self._make_request(
            "updateUser",
            {"username": "bob", "email": "email@email.em", "jukeboxRole": True},
        )
        rv, child = self._make_request("getUser", {"username": "bob"}, tag="user")
        self.assertEqual(child.get("email"), "email@email.em")
        self.assertEqual(child.get("adminRole"), "false")
        self.assertEqual(child.get("jukeboxRole"), "true")

        self._make_request(
            "updateUser", {"username": "bob", "email": "example@email.com"}
        )
        rv, child = self._make_request("getUser", {"username": "bob"}, tag="user")
        self.assertEqual(child.get("email"), "example@email.com")
        self.assertEqual(child.get("adminRole"), "false")
        self.assertEqual(child.get("jukeboxRole"), "true")

        self._make_request("updateUser", {"username": "bob", "adminRole": True})
        rv, child = self._make_request("getUser", {"username": "bob"}, tag="user")
        self.assertEqual(child.get("email"), "example@email.com")
        self.assertEqual(child.get("adminRole"), "true")
        self.assertEqual(child.get("jukeboxRole"), "true")


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2019-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import logging
import time

from multiprocessing.connection import Listener, Client
from threading import Thread, Event

from .client import DaemonCommand
from ..db import Folder, open_connection, close_connection
from ..jukebox import Jukebox
from ..scanner import Scanner
from ..utils import get_secret_key
from ..watcher import SupysonicWatcher

__all__ = ["Daemon"]

logger = logging.getLogger(__name__)


class Daemon:
    def __init__(self, config):
        self.__config = config
        self.__listener = None
        self.__watcher = None
        self.__scanner = None
        self.__jukebox = None
        self.__stopped = Event()

    watcher = property(lambda self: self.__watcher)
    scanner = property(lambda self: self.__scanner)
    jukebox = property(lambda self: self.__jukebox)

    def __handle_connection(self, connection):
        cmd = connection.recv()
        logger.debug("Received %s", cmd)
        if cmd is None:
            pass
        elif isinstance(cmd, DaemonCommand):
            cmd.apply(connection, self)
        else:
            logger.warn("Received unknown command %s", cmd)

    def run(self):
        self.__listener = Listener(
            address=self.__config.DAEMON["socket"], authkey=get_secret_key("daemon_key")
        )
        logger.info("Listening to %s", self.__listener.address)

        if self.__config.DAEMON["run_watcher"]:
            self.__watcher = SupysonicWatcher(self.__config)
            self.__watcher.start()

        if self.__config.DAEMON["jukebox_command"]:
            self.__jukebox = Jukebox(self.__config.DAEMON["jukebox_command"])

        close_connection()

        Thread(target=self.__listen).start()
        while not self.__stopped.is_set():
            time.sleep(1)

    def __listen(self):
        while not self.__stopped.is_set():
            conn = self.__listener.accept()
            self.__handle_connection(conn)

        self.__listener.close()

    def start_scan(self, folders=[], force=False):
        if not folders:
            open_connection()
            folders = [
                t[0] for t in Folder.select(Folder.name).where(Folder.root).tuples()
            ]
            close_connection()

        if self.__scanner is not None and self.__scanner.is_alive():
            for f in folders:
                self.__scanner.queue_folder(f)
            return

        extensions = self.__config.BASE["scanner_extensions"]
        if extensions:
            extensions = extensions.split(" ")

        self.__scanner = Scanner(
            force=force,
            extensions=extensions,
            follow_symlinks=self.__config.BASE["follow_symlinks"],
            on_folder_start=self.__unwatch,
            on_folder_end=self.__watch,
        )
        for f in folders:
            self.__scanner.queue_folder(f)

        self.__scanner.start()

    def __watch(self, folder):
        if self.__watcher is not None:
            self.__watcher.add_folder(folder.path)

    def __unwatch(self, folder):
        if self.__watcher is not None:
            self.__watcher.remove_folder(folder.path)

    def terminate(self):
        with Client(self.__listener.address, authkey=self.__listener._authkey) as c:
            self.__stopped.set()
            c.send(None)

        if self.__scanner is not None:
            self.__scanner.stop()
            self.__scanner.join()
        if self.__watcher is not None:
            self.__watcher.stop()
        if self.__jukebox is not None:
            self.__jukebox.terminate()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2019-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import shutil
import tempfile
import unittest

from supysonic.db import init_database, release_database
from supysonic.db import Track
from supysonic.managers.folder import FolderManager
from supysonic.scanner import Scanner


class Issue133TestCase(unittest.TestCase):
    def setUp(self):
        self.__dir = tempfile.mkdtemp()
        shutil.copy("tests/assets/issue133.flac", self.__dir)
        init_database("sqlite:")
        FolderManager.add("folder", self.__dir)

    def tearDown(self):
        release_database()
        shutil.rmtree(self.__dir)

    def test_issue133(self):
        scanner = Scanner()
        scanner.queue_folder("folder")
        scanner.run()

        track = Track.select().first()
        self.assertNotIn("\x00", track.title)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2018-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import os.path
import unittest

from supysonic.db import User, Track, StarredTrack, RatingTrack
from supysonic.managers.folder import FolderManager
from supysonic.scanner import Scanner

from .testbase import TestBase


class Issue129TestCase(TestBase):
    def setUp(self):
        super().setUp()

        FolderManager.add("folder", os.path.abspath("tests/assets/folder"))
        scanner = Scanner()
        scanner.queue_folder("folder")
        scanner.run()

        self.trackid = Track.select().first().id
        self.userid = User.get(name="alice").id

    def test_last_play(self):
        user = User[self.userid]
        user.last_play = Track[self.trackid]
        user.save()
        FolderManager.delete_by_name("folder")

    def test_starred(self):
        StarredTrack.create(user=self.userid, starred=self.trackid)
        FolderManager.delete_by_name("folder")

    def test_rating(self):
        RatingTrack.create(user=self.userid, rated=self.trackid, rating=5)
        FolderManager.delete_by_name("folder")


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2018 Alban 'spl0k' Féron
#               2018-2019 Carey 'pR0Ps' Metcalfe
#
# Distributed under terms of the GNU AGPLv3 license.

import os
import unittest
import shutil
import time
import tempfile

from supysonic.cache import Cache, CacheMiss, ProtectedError


class CacheTestCase(unittest.TestCase):
    def setUp(self):
        self.__dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.__dir)

    def test_existing_files_order(self):
        cache = Cache(self.__dir, 30)
        val = b"0123456789"
        cache.set("key1", val)
        cache.set("key2", val)
        cache.set("key3", val)
        self.assertEqual(cache.size, 30)

        # file mtime is accurate to the second
        time.sleep(1)
        cache.get_value("key1")

        cache = Cache(self.__dir, 30, min_time=0)
        self.assertEqual(cache.size, 30)
        self.assertTrue(cache.has("key1"))
        self.assertTrue(cache.has("key2"))
        self.assertTrue(cache.has("key3"))

        cache.set("key4", val)
        self.assertEqual(cache.size, 30)
        self.assertTrue(cache.has("key1"))
        self.assertFalse(cache.has("key2"))
        self.assertTrue(cache.has("key3"))
        self.assertTrue(cache.has("key4"))

    def test_missing(self):
        cache = Cache(self.__dir, 10)
        self.assertFalse(cache.has("missing"))
        with self.assertRaises(CacheMiss):
            cache.get_value("missing")

    def test_delete_missing(self):
        cache = Cache(self.__dir, 0, min_time=0)
        cache.delete("missing1")
        cache.delete("missing2")

    def test_store_literal(self):
        cache = Cache(self.__dir, 10)
        val = b"0123456789"
        cache.set("key", val)
        self.assertEqual(cache.size, 10)
        self.assertTrue(cache.has("key"))
        self.assertEqual(cache.get_value("key"), val)

    def test_store_generated(self):
        cache = Cache(self.__dir, 10)
        val = [b"0", b"12", b"345", b"6789"]

        def gen():
            yield from val

        t = []
        for x in cache.set_generated("key", gen):
            t.append(x)
            self.assertEqual(cache.size, 0)
            self.assertFalse(cache.has("key"))

        self.assertEqual(t, val)
        self.assertEqual(cache.size, 10)
        self.assertEqual(cache.get_value("key"), b"".join(val))

    def test_store_to_fp(self):
        cache = Cache(self.__dir, 10)
        val = b"0123456789"
        with cache.set_fileobj("key") as fp:
            fp.write(val)
            self.assertEqual(cache.size, 0)

        self.assertEqual(cache.size, 10)
        self.assertEqual(cache.get_value("key"), val)

    def test_access_data(self):
        cache = Cache(self.__dir, 25, min_time=0)
        val = b"0123456789"
        cache.set("key", val)

        self.assertEqual(cache.get_value("key"), val)

        with cache.get_fileobj("key") as f:
            self.assertEqual(f.read(), val)

        with open(cache.get("key"), "rb") as f:
            self.assertEqual(f.read(), val)

    def test_accessing_preserves(self):
        cache = Cache(self.__dir, 25, min_time=0)
        val = b"0123456789"
        cache.set("key1", val)
        cache.set("key2", val)
        self.assertEqual(cache.size, 20)

        cache.get_value("key1")

        cache.set("key3", val)
        self.assertEqual(cache.size, 20)
        self.assertTrue(cache.has("key1"))
        self.assertFalse(cache.has("key2"))
        self.assertTrue(cache.has("key3"))

    def test_automatic_delete_oldest(self):
        cache = Cache(self.__dir, 25, min_time=0)
        val = b"0123456789"
        cache.set("key1", val)
        self.assertTrue(cache.has("key1"))
        self.assertEqual(cache.size, 10)

        cache.set("key2", val)
        self.assertEqual(cache.size, 20)
        self.assertTrue(cache.has("key1"))
        self.assertTrue(cache.has("key2"))

        cache.set("key3", val)
        self.assertEqual(cache.size, 20)
        self.assertFalse(cache.has("key1"))
        self.assertTrue(cache.has("key2"))
        self.assertTrue(cache.has("key3"))

    def test_delete(self):
        cache = Cache(self.__dir, 25, min_time=0)
        val = b"0123456789"
        cache.set("key1", val)
        self.assertTrue(cache.has("key1"))
        self.assertEqual(cache.size, 10)

        cache.delete("key1")

        self.assertFalse(cache.has("key1"))
        self.assertEqual(cache.size, 0)

    def test_cleanup_on_error(self):
        cache = Cache(self.__dir, 10)

        def gen():
            # Cause a TypeError halfway through
            yield from [b"0", b"12", object(), b"345", b"6789"]

        with self.assertRaises(TypeError):
            for x in cache.set_generated("key", gen):
                pass

        # Make sure no partial files are left after the error
        self.assertEqual(list(os.listdir(self.__dir)), list())

    def test_parallel_generation(self):
        cache = Cache(self.__dir, 20)

        def gen():
            yield from [b"0", b"12", b"345", b"6789"]

        g1 = cache.set_generated("key", gen)
        g2 = cache.set_generated("key", gen)

        next(g1)
        files = os.listdir(self.__dir)
        self.assertEqual(len(files), 1)
        for x in files:
            self.assertTrue(x.endswith(".part"))

        next(g2)
        files = os.listdir(self.__dir)
        self.assertEqual(len(files), 2)
        for x in files:
            self.assertTrue(x.endswith(".part"))

        self.assertEqual(cache.size, 0)
        for x in g1:
            pass
        self.assertEqual(cache.size, 10)
        self.assertTrue(cache.has("key"))

        # Replace the file - size should stay the same
        for x in g2:
            pass
        self.assertEqual(cache.size, 10)
        self.assertTrue(cache.has("key"))

        # Only a single file
        self.assertEqual(len(os.listdir(self.__dir)), 1)

    def test_replace(self):
        cache = Cache(self.__dir, 20)
        val_small = b"0"
        val_big = b"0123456789"

        cache.set("key", val_small)
        self.assertEqual(cache.size, 1)

        cache.set("key", val_big)
        self.assertEqual(cache.size, 10)

        cache.set("key", val_small)
        self.assertEqual(cache.size, 1)

    def test_no_auto_prune(self):
        cache = Cache(self.__dir, 10, min_time=0, auto_prune=False)
        val = b"0123456789"

        cache.set("key1", val)
        cache.set("key2", val)
        cache.set("key3", val)
        cache.set("key4", val)
        self.assertEqual(cache.size, 40)
        cache.prune()

        self.assertEqual(cache.size, 10)

    def test_min_time_clear(self):
        cache = Cache(self.__dir, 40, min_time=1)
        val = b"0123456789"

        cache.set("key1", val)
        cache.set("key2", val)
        time.sleep(1)
        cache.set("key3", val)
        cache.set("key4", val)

        self.assertEqual(cache.size, 40)
        cache.clear()
        self.assertEqual(cache.size, 20)
        time.sleep(1)
        cache.clear()
        self.assertEqual(cache.size, 0)

    def test_not_expired(self):
        cache = Cache(self.__dir, 40, min_time=1)
        val = b"0123456789"
        cache.set("key1", val)
        with self.assertRaises(ProtectedError):
            cache.delete("key1")
        time.sleep(1)
        cache.delete("key1")
        self.assertEqual(cache.size, 0)

    def test_missing_cache_file(self):
        cache = Cache(self.__dir, 10, min_time=0)
        val = b"0123456789"
        os.remove(cache.set("key", val))

        self.assertEqual(cache.size, 10)
        self.assertFalse(cache.has("key"))
        self.assertEqual(cache.size, 0)

        os.remove(cache.set("key", val))
        self.assertEqual(cache.size, 10)
        with self.assertRaises(CacheMiss):
            cache.get("key")
        self.assertEqual(cache.size, 0)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2021-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from waitress import serve

from ._base import BaseServer


class WaitressServer(BaseServer):
    def _build_kwargs(self):
        rv = {"app": self._load_app()}

        if self._host is not None:
            rv["host"] = self._host
        if self._port is not None:
            rv["port"] = self._port
        if self._socket is not None:
            rv["unix_socket"] = self._socket
        if self._threads is not None:
            rv["threads"] = self._threads

        return rv

    def _run(self, **kwargs):
        return serve(**kwargs)


server = WaitressServer



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2021-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest

from supysonic import db


class Issue221TestCase(unittest.TestCase):
    def setUp(self):
        db.init_database("sqlite:")
        root = db.Folder.create(root=True, name="Folder", path="tests")
        artist = db.Artist.create(name="Artist")
        album = db.Album.create(artist=artist, name="Album")

        for i in range(3):
            db.Track.create(
                title=f"Track {i}",
                album=album,
                artist=artist,
                disc=1,
                number=i + 1,
                duration=3,
                has_art=False,
                bitrate=64,
                path=f"tests/track{i}",
                last_modification=2,
                root_folder=root,
                folder=root,
                genre="Genre",
            )

        db.User.create(name="user", password="secret", salt="sugar")

    def tearDown(self):
        db.release_database()

    def test_issue(self):
        data = db.Album.get().as_subsonic_album(db.User.get())
        self.assertIn("genre", data)
        self.assertEqual(data["genre"], "Genre")


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2020 Alban 'spl0k' Féron
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

import os
import sys
import tempfile

from configparser import RawConfigParser

current_config = None


def get_current_config():
    return current_config or DefaultConfig()


class DefaultConfig:
    DEBUG = False

    tempdir = os.path.join(tempfile.gettempdir(), "supysonic")
    BASE = {
        "database_uri": "sqlite:///" + os.path.join(tempdir, "supysonic.db"),
        "scanner_extensions": None,
        "follow_symlinks": False,
    }
    WEBAPP = {
        "cache_dir": tempdir,
        "cache_size": 1024,
        "transcode_cache_size": 512,
        "log_file": None,
        "log_level": "WARNING",
        "log_rotate": True,
        "mount_webui": True,
        "mount_api": True,
        "index_ignored_prefixes": "El La Le Las Les Los The",
        "online_lyrics": False,
    }
    DAEMON = {
        "socket": (
            r"\\.\pipe\supysonic"
            if sys.platform == "win32"
            else os.path.join(tempdir, "supysonic.sock")
        ),
        "run_watcher": True,
        "wait_delay": 5,
        "jukebox_command": None,
        "log_file": None,
        "log_level": "WARNING",
        "log_rotate": True,
    }
    LASTFM = {"api_key": None, "secret": None}
    LISTENBRAINZ = {"api_url": "https://api.listenbrainz.org"}
    TRANSCODING = {}
    MIMETYPES = {}

    def __init__(self):
        current_config = self


class IniConfig(DefaultConfig):
    common_paths = [
        "/etc/supysonic",
        os.path.expanduser("~/.supysonic"),
        os.path.expanduser("~/.config/supysonic/supysonic.conf"),
        "supysonic.conf",
    ]

    def __init__(self, paths):
        super().__init__()

        parser = RawConfigParser()
        parser.read(paths)

        for section in parser.sections():
            options = {k: self.__try_parse(v) for k, v in parser.items(section)}
            section = section.upper()

            if hasattr(self, section):
                getattr(self, section).update(options)
            else:
                setattr(self, section, options)

    @staticmethod
    def __try_parse(value):
        try:
            return int(value)
        except ValueError:
            try:
                return float(value)
            except ValueError:
                lv = value.lower()
                if lv in ("yes", "true", "on"):
                    return True
                if lv in ("no", "false", "off"):
                    return False
                return value

    @classmethod
    def from_common_locations(cls):
        return IniConfig(cls.common_paths)



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2020 Alban 'spl0k' Féron
#               2020 Vincent Ducamps
#
# Distributed under terms of the GNU AGPLv3 license.

from flask import request
from flask import current_app

from ..daemon.client import DaemonClient
from ..daemon.exceptions import DaemonUnavailableError

from . import api_routing
from .user import admin_only
from .exceptions import ServerError


@api_routing("/startScan")
@admin_only
def startScan():
    try:
        daemonclient = DaemonClient(current_app.config["DAEMON"]["socket"])
        daemonclient.scan()
        scanned = daemonclient.get_scanning_progress()
    except DaemonUnavailableError as e:
        raise ServerError(str(e))
    return request.formatter(
        "scanStatus",
        {
            "scanning": scanned is not None,
            "count": scanned or 0,
        },
    )


@api_routing("/getScanStatus")
@admin_only
def getScanStatus():
    try:
        scanned = DaemonClient(
            current_app.config["DAEMON"]["socket"]
        ).get_scanning_progress()
    except DaemonUnavailableError as e:
        raise ServerError(str(e))
    return request.formatter(
        "scanStatus",
        {
            "scanning": scanned is not None,
            "count": scanned or 0,
        },
    )



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2018-2020 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from flask import current_app, request
from werkzeug.exceptions import HTTPException


class SubsonicAPIException(HTTPException):
    code = 400
    api_code = None
    message = None

    def get_response(self, environ=None):
        rv = request.formatter.error(self.api_code, self.message)
        # rv.status_code = self.code
        return rv

    def __str__(self):
        code = self.api_code if self.api_code is not None else "??"
        return f"{code}: {self.message}"


class GenericError(SubsonicAPIException):
    api_code = 0

    def __init__(self, message, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message = message


class ServerError(GenericError):
    code = 500


class UnsupportedParameter(GenericError):
    def __init__(self, parameter, *args, **kwargs):
        message = f"Unsupported parameter '{parameter}'"
        super().__init__(message, *args, **kwargs)


class MissingParameter(SubsonicAPIException):
    api_code = 10

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message = "A required parameter is missing."


class ClientMustUpgrade(SubsonicAPIException):
    api_code = 20
    message = "Incompatible Subsonic REST protocol version. Client must upgrade."


class ServerMustUpgrade(SubsonicAPIException):
    code = 501
    api_code = 30
    message = "Incompatible Subsonic REST protocol version. Server must upgrade."


class Unauthorized(SubsonicAPIException):
    code = 401
    api_code = 40
    message = "Wrong username or password."


class Forbidden(SubsonicAPIException):
    code = 403
    api_code = 50
    message = "User is not authorized for the given operation."


class TrialExpired(SubsonicAPIException):
    code = 402
    api_code = 60
    message = (
        "The trial period for the Supysonic server is over."
        "But since it doesn't use any licensing you shouldn't be seeing this error ever."
        "So something went wrong or you got scammed."
    )


class NotFound(SubsonicAPIException):
    code = 404
    api_code = 70

    def __init__(self, entity, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message = f"{entity} not found"


class AggregateException(SubsonicAPIException):
    def __init__(self, exceptions, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.exceptions = []
        for exc in exceptions:
            if not isinstance(exc, SubsonicAPIException):
                # Try to convert regular exceptions to SubsonicAPIExceptions
                handler = current_app._find_error_handler(exc)  # meh
                if handler:
                    exc = handler(exc)
                    assert isinstance(exc, SubsonicAPIException)
                else:
                    exc = GenericError(str(exc))
            self.exceptions.append(exc)

    def get_response(self, environ=None):
        if len(self.exceptions) == 1:
            return self.exceptions[0].get_response()

        codes = {exc.api_code for exc in self.exceptions}
        errors = [
            {"code": exc.api_code, "message": exc.message} for exc in self.exceptions
        ]

        rv = request.formatter(
            "error",
            {"code": next(iter(codes)) if len(codes) == 1 else 0, "error": errors},
        )
        # rv.status_code = self.code
        return rv



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import time
import unittest

from supysonic.db import Folder, Artist, Album, Track

from .apitestbase import ApiTestBase


class SearchTestCase(ApiTestBase):
    def setUp(self):
        super().setUp()

        root = Folder.create(root=True, name="Root folder", path="tests/assets")
        Folder.create(root=True, name="Empty", path="/tmp")

        for letter in "ABC":
            folder = Folder.create(
                name=letter + "rtist",
                path=f"tests/assets/{letter}rtist",
                root=False,
                parent=root,
            )
            artist = Artist.create(name=letter + "rtist")

            for lether in "AB":
                afolder = Folder.create(
                    name=letter + lether + "lbum",
                    path="tests/assets/{0}rtist/{0}{1}lbum".format(letter, lether),
                    root=False,
                    parent=folder,
                )

                album = Album.create(name=letter + lether + "lbum", artist=artist)

                for num, song in enumerate(["One", "Two", "Three"]):
                    Track.create(
                        disc=1,
                        number=num,
                        title=song,
                        duration=2,
                        album=album,
                        artist=artist,
                        bitrate=320,
                        path="tests/assets/{0}rtist/{0}{1}lbum/{2}".format(
                            letter, lether, song
                        ),
                        last_modification=0,
                        root_folder=root,
                        folder=afolder,
                    )

        self.assertEqual(Folder.select().count(), 11)
        self.assertEqual(Artist.select().count(), 3)
        self.assertEqual(Album.select().count(), 6)
        self.assertEqual(Track.select().count(), 18)

    def __track_as_pseudo_unique_str(self, elem):
        return elem.get("artist") + elem.get("album") + elem.get("title")

    def test_search(self):
        # invalid parameters
        self._make_request("search", {"count": "string"}, error=0)
        self._make_request("search", {"offset": "sstring"}, error=0)
        self._make_request("search", {"newerThan": "ssstring"}, error=0)

        # no search
        self._make_request("search", error=10)

        # non existent artist (but searched string present in other fields)
        rv, child = self._make_request("search", {"artist": "One"}, tag="searchResult")
        self.assertEqual(len(child), 0)
        self.assertEqual(child.get("totalHits"), "0")
        self.assertEqual(child.get("offset"), "0")

        rv, child = self._make_request(
            "search", {"artist": "AAlbum"}, tag="searchResult"
        )
        self.assertEqual(len(child), 0)

        # non existent album (but search present in other fields)
        rv, child = self._make_request("search", {"album": "One"}, tag="searchResult")
        self.assertEqual(len(child), 0)

        rv, child = self._make_request(
            "search", {"album": "Artist"}, tag="searchResult"
        )
        self.assertEqual(len(child), 0)

        # non existent title (but search present in other fields)
        rv, child = self._make_request(
            "search", {"title": "AAlbum"}, tag="searchResult"
        )
        self.assertEqual(len(child), 0)

        rv, child = self._make_request(
            "search", {"title": "Artist"}, tag="searchResult"
        )
        self.assertEqual(len(child), 0)

        # non existent anything
        rv, child = self._make_request("search", {"any": "Chaos"}, tag="searchResult")
        self.assertEqual(len(child), 0)

        # artist search
        rv, child = self._make_request(
            "search", {"artist": "Artist"}, tag="searchResult"
        )
        self.assertEqual(len(child), 1)
        self.assertEqual(child.get("totalHits"), "1")
        self.assertEqual(child[0].get("title"), "Artist")

        rv, child = self._make_request("search", {"artist": "rti"}, tag="searchResult")
        self.assertEqual(len(child), 3)
        self.assertEqual(child.get("totalHits"), "3")

        # same as above, but created in the future
        future = int(time.time() * 1000 + 1000)
        rv, child = self._make_request(
            "search", {"artist": "rti", "newerThan": future}, tag="searchResult"
        )
        self.assertEqual(len(child), 0)

        # album search
        rv, child = self._make_request(
            "search", {"album": "AAlbum"}, tag="searchResult"
        )
        self.assertEqual(len(child), 1)
        self.assertEqual(child[0].get("title"), "AAlbum")
        self.assertEqual(child[0].get("artist"), "Artist")

        rv, child = self._make_request("search", {"album": "lbu"}, tag="searchResult")
        self.assertEqual(len(child), 6)

        # same as above, but created in the future
        rv, child = self._make_request(
            "search", {"album": "lbu", "newerThan": future}, tag="searchResult"
        )
        self.assertEqual(len(child), 0)

        # song search
        rv, child = self._make_request("search", {"title": "One"}, tag="searchResult")
        self.assertEqual(len(child), 6)
        for i in range(6):
            self.assertEqual(child[i].get("title"), "One")

        rv, child = self._make_request("search", {"title": "e"}, tag="searchResult")
        self.assertEqual(len(child), 12)

        # same as above, but created in the future
        rv, child = self._make_request(
            "search", {"title": "e", "newerThan": future}, tag="searchResult"
        )
        self.assertEqual(len(child), 0)

        # any field search
        rv, child = self._make_request("search", {"any": "r"}, tag="searchResult")
        self.assertEqual(len(child), 10)  # root + 3 artists (*rtist)  + 6 songs (Three)

        # same as above, but created in the future
        rv, child = self._make_request(
            "search", {"any": "r", "newerThan": future}, tag="searchResult"
        )
        self.assertEqual(len(child), 0)

        # paging
        songs = []
        for offset in range(0, 12, 2):
            rv, child = self._make_request(
                "search",
                {"title": "e", "count": 2, "offset": offset},
                tag="searchResult",
            )
            self.assertEqual(len(child), 2)
            self.assertEqual(child.get("totalHits"), "12")
            self.assertEqual(child.get("offset"), str(offset))
            for song in map(self.__track_as_pseudo_unique_str, child):
                self.assertNotIn(song, songs)
                songs.append(song)

    def test_search2(self):
        # invalid parameters
        self._make_request("search2", {"query": "a", "artistCount": "string"}, error=0)
        self._make_request(
            "search2", {"query": "a", "artistOffset": "sstring"}, error=0
        )
        self._make_request("search2", {"query": "a", "albumCount": "string"}, error=0)
        self._make_request("search2", {"query": "a", "albumOffset": "sstring"}, error=0)
        self._make_request("search2", {"query": "a", "songCount": "string"}, error=0)
        self._make_request("search2", {"query": "a", "songOffset": "sstring"}, error=0)
        self._make_request(
            "search2", {"query": "a", "musicFolderId": "sstring"}, error=0
        )
        self._make_request("search2", {"query": "a", "musicFolderId": -2}, error=70)

        # no search
        self._make_request("search2", error=10)

        # non existent anything
        rv, child = self._make_request(
            "search2", {"query": "Chaos"}, tag="searchResult2"
        )
        self.assertEqual(len(child), 0)

        # artist search
        rv, child = self._make_request(
            "search2", {"query": "Artist"}, tag="searchResult2"
        )
        self.assertEqual(len(child), 1)
        self.assertEqual(len(self._xpath(child, "./artist")), 1)
        self.assertEqual(len(self._xpath(child, "./album")), 0)
        self.assertEqual(len(self._xpath(child, "./song")), 0)
        self.assertEqual(child[0].get("name"), "Artist")

        rv, child = self._make_request("search2", {"query": "rti"}, tag="searchResult2")
        self.assertEqual(len(child), 3)
        self.assertEqual(len(self._xpath(child, "./artist")), 3)
        self.assertEqual(len(self._xpath(child, "./album")), 0)
        self.assertEqual(len(self._xpath(child, "./song")), 0)

        # album search
        rv, child = self._make_request(
            "search2", {"query": "AAlbum"}, tag="searchResult2"
        )
        self.assertEqual(len(child), 1)
        self.assertEqual(len(self._xpath(child, "./artist")), 0)
        self.assertEqual(len(self._xpath(child, "./album")), 1)
        self.assertEqual(len(self._xpath(child, "./song")), 0)
        self.assertEqual(child[0].get("title"), "AAlbum")
        self.assertEqual(child[0].get("artist"), "Artist")

        rv, child = self._make_request("search2", {"query": "lbu"}, tag="searchResult2")
        self.assertEqual(len(child), 6)
        self.assertEqual(len(self._xpath(child, "./artist")), 0)
        self.assertEqual(len(self._xpath(child, "./album")), 6)
        self.assertEqual(len(self._xpath(child, "./song")), 0)

        # song search
        rv, child = self._make_request("search2", {"query": "One"}, tag="searchResult2")
        self.assertEqual(len(child), 6)
        self.assertEqual(len(self._xpath(child, "./artist")), 0)
        self.assertEqual(len(self._xpath(child, "./album")), 0)
        self.assertEqual(len(self._xpath(child, "./song")), 6)
        for i in range(6):
            self.assertEqual(child[i].get("title"), "One")

        rv, child = self._make_request("search2", {"query": "e"}, tag="searchResult2")
        self.assertEqual(len(child), 12)
        self.assertEqual(len(self._xpath(child, "./artist")), 0)
        self.assertEqual(len(self._xpath(child, "./album")), 0)
        self.assertEqual(len(self._xpath(child, "./song")), 12)

        # any field search
        rv, child = self._make_request("search2", {"query": "r"}, tag="searchResult2")
        self.assertEqual(len(child), 9)
        self.assertEqual(len(self._xpath(child, "./artist")), 3)
        self.assertEqual(len(self._xpath(child, "./album")), 0)
        self.assertEqual(len(self._xpath(child, "./song")), 6)

        # paging
        artists = []
        for offset in range(0, 4, 2):
            rv, child = self._make_request(
                "search2",
                {"query": "r", "artistCount": 2, "artistOffset": offset},
                tag="searchResult2",
            )
            names = self._xpath(child, "./artist/@name")
            self.assertLessEqual(len(names), 2)
            for name in names:
                self.assertNotIn(name, artists)
                artists.append(name)

        songs = []
        for offset in range(0, 6, 2):
            rv, child = self._make_request(
                "search2",
                {"query": "r", "songCount": 2, "songOffset": offset},
                tag="searchResult2",
            )
            elems = self._xpath(child, "./song")
            self.assertEqual(len(elems), 2)
            for song in map(self.__track_as_pseudo_unique_str, elems):
                self.assertNotIn(song, songs)
                songs.append(song)

        # root filtering
        _, child = self._make_request(
            "search2", {"query": "One", "musicFolderId": 1}, tag="searchResult2"
        )
        self.assertEqual(len(self._xpath(child, "./song")), 6)

        _, child = self._make_request(
            "search2", {"query": "One", "musicFolderId": 2}, tag="searchResult2"
        )
        self.assertEqual(len(self._xpath(child, "./song")), 0)

    # Almost identical as above. Test dataset (and tests) should probably be changed
    # to have folders that don't share names with artists or albums
    def test_search3(self):
        # invalid parameters
        self._make_request("search3", {"query": "a", "artistCount": "string"}, error=0)
        self._make_request(
            "search3", {"query": "a", "artistOffset": "sstring"}, error=0
        )
        self._make_request("search3", {"query": "a", "albumCount": "string"}, error=0)
        self._make_request("search3", {"query": "a", "albumOffset": "sstring"}, error=0)
        self._make_request("search3", {"query": "a", "songCount": "string"}, error=0)
        self._make_request("search3", {"query": "a", "songOffset": "sstring"}, error=0)
        self._make_request(
            "search3", {"query": "a", "musicFolderId": "sstring"}, error=0
        )
        self._make_request("search3", {"query": "a", "musicFolderId": -2}, error=70)

        # no search
        self._make_request("search3", error=10)

        # non existent anything
        rv, child = self._make_request(
            "search3", {"query": "Chaos"}, tag="searchResult3"
        )
        self.assertEqual(len(child), 0)

        # artist search
        rv, child = self._make_request(
            "search3", {"query": "Artist"}, tag="searchResult3"
        )
        self.assertEqual(len(child), 1)
        self.assertEqual(len(self._xpath(child, "./artist")), 1)
        self.assertEqual(len(self._xpath(child, "./album")), 0)
        self.assertEqual(len(self._xpath(child, "./song")), 0)
        self.assertEqual(child[0].get("name"), "Artist")

        rv, child = self._make_request("search3", {"query": "rti"}, tag="searchResult3")
        self.assertEqual(len(child), 3)
        self.assertEqual(len(self._xpath(child, "./artist")), 3)
        self.assertEqual(len(self._xpath(child, "./album")), 0)
        self.assertEqual(len(self._xpath(child, "./song")), 0)

        # album search
        rv, child = self._make_request(
            "search3", {"query": "AAlbum"}, tag="searchResult3"
        )
        self.assertEqual(len(child), 1)
        self.assertEqual(len(self._xpath(child, "./artist")), 0)
        self.assertEqual(len(self._xpath(child, "./album")), 1)
        self.assertEqual(len(self._xpath(child, "./song")), 0)
        self.assertEqual(child[0].get("name"), "AAlbum")
        self.assertEqual(child[0].get("artist"), "Artist")

        rv, child = self._make_request("search3", {"query": "lbu"}, tag="searchResult3")
        self.assertEqual(len(child), 6)
        self.assertEqual(len(self._xpath(child, "./artist")), 0)
        self.assertEqual(len(self._xpath(child, "./album")), 6)
        self.assertEqual(len(self._xpath(child, "./song")), 0)

        # song search
        rv, child = self._make_request("search3", {"query": "One"}, tag="searchResult3")
        self.assertEqual(len(child), 6)
        self.assertEqual(len(self._xpath(child, "./artist")), 0)
        self.assertEqual(len(self._xpath(child, "./album")), 0)
        self.assertEqual(len(self._xpath(child, "./song")), 6)
        for i in range(6):
            self.assertEqual(child[i].get("title"), "One")

        rv, child = self._make_request("search3", {"query": "e"}, tag="searchResult3")
        self.assertEqual(len(child), 12)
        self.assertEqual(len(self._xpath(child, "./artist")), 0)
        self.assertEqual(len(self._xpath(child, "./album")), 0)
        self.assertEqual(len(self._xpath(child, "./song")), 12)

        # any field search
        rv, child = self._make_request("search3", {"query": "r"}, tag="searchResult3")
        self.assertEqual(len(child), 9)
        self.assertEqual(len(self._xpath(child, "./artist")), 3)
        self.assertEqual(len(self._xpath(child, "./album")), 0)
        self.assertEqual(len(self._xpath(child, "./song")), 6)

        # paging
        artists = []
        for offset in range(0, 4, 2):
            rv, child = self._make_request(
                "search3",
                {"query": "r", "artistCount": 2, "artistOffset": offset},
                tag="searchResult3",
            )
            names = self._xpath(child, "./artist/@name")
            self.assertLessEqual(len(names), 2)
            for name in names:
                self.assertNotIn(name, artists)
                artists.append(name)

        songs = []
        for offset in range(0, 6, 2):
            rv, child = self._make_request(
                "search3",
                {"query": "r", "songCount": 2, "songOffset": offset},
                tag="searchResult3",
            )
            elems = self._xpath(child, "./song")
            self.assertEqual(len(elems), 2)
            for song in map(self.__track_as_pseudo_unique_str, elems):
                self.assertNotIn(song, songs)
                songs.append(song)

        # root filtering
        _, child = self._make_request(
            "search3", {"query": "One", "musicFolderId": 1}, tag="searchResult3"
        )
        self.assertEqual(len(self._xpath(child, "./song")), 6)

        _, child = self._make_request(
            "search3", {"query": "One", "musicFolderId": 2}, tag="searchResult3"
        )
        self.assertEqual(len(self._xpath(child, "./song")), 0)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2021-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import os
import os.path

from gevent import socket
from gevent.pywsgi import WSGIServer

from ._base import BaseServer


class GeventServer(BaseServer):
    def _build_kwargs(self):
        rv = {"application": self._load_app()}

        if self._socket is not None:
            if os.path.exists(self._socket):
                os.remove(self._socket)

            listener = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            listener.bind(self._socket)
            listener.listen()

            rv["listener"] = listener
        else:
            rv["listener"] = (self._host, self._port)

        return rv

    def _run(self, **kwargs):
        return WSGIServer(**kwargs).serve_forever()


server = GeventServer



import argparse
import hashlib
import psycopg2

try:
    bytes = buffer
except NameError:
    pass

parser = argparse.ArgumentParser()
parser.add_argument("username")
parser.add_argument("password")
parser.add_argument("database")
parser.add_argument("-H", "--host", default="localhost", help="default: localhost")
args = parser.parse_args()


def process_table(connection, table):
    c = connection.cursor()

    c.execute(
        r"ALTER TABLE {} ADD COLUMN path_hash BYTEA NOT NULL DEFAULT E'\\0000'".format(
            table
        )
    )

    hashes = dict()
    c.execute("SELECT path FROM {}".format(table))
    for row in c.fetchall():
        hashes[row[0]] = hashlib.sha1(row[0].encode("utf-8")).digest()
    c.executemany(
        "UPDATE {} SET path_hash=%s WHERE path=%s".format(table),
        [(bytes(h), p) for p, h in hashes.items()],
    )

    c.execute("CREATE UNIQUE INDEX index_{0}_path ON {0}(path_hash)".format(table))


with psycopg2.connect(
    host=args.host, user=args.username, password=args.password, dbname=args.database
) as conn:
    process_table(conn, "folder")
    process_table(conn, "track")



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from flask import request

from ..db import ChatMessage
from . import api_routing


@api_routing("/getChatMessages")
def get_chat():
    since = request.values.get("since")
    since = int(since) / 1000 if since else None

    query = ChatMessage.select().order_by(ChatMessage.time)
    if since:
        query = query.where(ChatMessage.time > since)

    return request.formatter(
        "chatMessages", {"chatMessage": [msg.responsize() for msg in query]}
    )


@api_routing("/addChatMessage")
def add_chat_message():
    msg = request.values["message"]
    ChatMessage.create(user=request.user, message=msg)

    return request.formatter.empty
